# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib.pyplot as plt
import matplotlib.tri as tri
import numpy as np

"""FOR DEVELOPMENT PURPOSES ONLY.

This module contains
"""

# TODO: use earthkit-plots


def fix(lons: np.ndarray) -> np.ndarray:
    """Fix longitudes greater than 180 degrees.

    Parameters
    ----------
    lons : np.ndarray
        Array of longitudes.

    Returns
    -------
    np.ndarray
        Array of fixed longitudes.
    """
    return np.where(lons > 180, lons - 360, lons)


def plot_values(
    values: np.ndarray,
    latitudes: np.ndarray,
    longitudes: np.ndarray,
    title: str = None,
    missing_value: float = None,
    min_value: float = None,
    max_value: float = None,
    **kwargs: dict,
) -> plt.Axes:
    """Plot values on a map.

    Parameters
    ----------
    values : np.ndarray
        Array of values to plot.
    latitudes : np.ndarray
        Array of latitudes.
    longitudes : np.ndarray
        Array of longitudes.
    title : str, optional
        Title of the plot, by default None.
    missing_value : float, optional
        Value to use for missing data, by default None.
    min_value : float, optional
        Minimum value for the plot, by default None.
    max_value : float, optional
        Maximum value for the plot, by default None.
    **kwargs : dict
        Additional keyword arguments for the plot.

    Returns
    -------
    plt.Axes
        The plot axes.
    """
    _, ax = plt.subplots(subplot_kw={"projection": ccrs.PlateCarree()})
    ax.coastlines()
    ax.add_feature(cfeature.BORDERS, linestyle=":")

    missing_values = np.isnan(values)
    if missing_value is None:
        min = np.nanmin(values)
        missing_value = min - np.abs(min) * 0.001

    values = np.where(missing_values, missing_value, values)

    if max_value is not None:
        values = np.where(values > max_value, max_value, values)

    if min_value is not None:
        values = np.where(values < min_value, min_value, values)

    triangulation = tri.Triangulation(fix(longitudes), latitudes)

    levels = kwargs.pop("levels", 10)

    _ = ax.tricontourf(triangulation, values, levels=levels, transform=ccrs.PlateCarree())

    options = dict(
        levels=levels,
        colors="black",
        linewidths=0.5,
        transform=ccrs.PlateCarree(),
    )

    options.update(kwargs)

    ax.tricontour(
        triangulation,
        values,
        **options,
    )

    if title is not None:
        ax.set_title(title)

    return ax


def plot_field(field: Any, title: str = None, **kwargs: dict) -> plt.Axes:
    """Plot a field on a map.

    Parameters
    ----------
    field : Any
        The field to plot.
    title : str, optional
        Title of the plot, by default None.
    **kwargs : dict
        Additional keyword arguments for the plot.

    Returns
    -------
    plt.Axes
        The plot axes.
    """
    values = field.to_numpy(flatten=True)
    latitudes, longitudes = field.grid_points()
    return plot_values(values, latitudes, longitudes, title=title, **kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
from textwrap import dedent

import yaml

from anemoi.utils.dates import datetimes_factory


def _(txt: str) -> datetimes_factory:
    """Helper function to create a datetimes_factory from a YAML string.

    Parameters
    ----------
    txt : str
        The YAML string.

    Returns
    -------
    datetimes_factory
        The datetimes_factory object.
    """
    txt = dedent(txt)
    config = yaml.safe_load(txt)
    return datetimes_factory(config)


def test_date_1() -> None:
    """Test datetimes_factory with a list of dates."""
    d = _(
        """
          - 2023-01-01
          - 2023-01-02
          - 2023-01-03
    """
    )
    assert len(list(d)) == 3


def test_date_2() -> None:
    """Test datetimes_factory with a date range and frequency."""
    d = _(
        """
        start: 2023-01-01
        end: 2023-01-07
        frequency: 12
        day_of_week: [monday, friday]
    """
    )
    assert len(list(d)) == 4


def test_date_3() -> None:
    """Test datetimes_factory with multiple date ranges and frequencies."""
    d = _(
        """
        - start: 2023-01-01
          end: 2023-01-03
          frequency: 24
        - start: 2024-01-01T06:00:00
          end: 2024-01-02T18:00:00
          frequency: 6h
    """
    )
    assert datetime.datetime(2023, 1, 1, 0) in d
    assert datetime.datetime(2023, 1, 2, 0) in d
    assert datetime.datetime(2023, 1, 3, 0) in d
    assert datetime.datetime(2024, 1, 1, 6) in d
    assert datetime.datetime(2024, 1, 1, 12) in d
    assert datetime.datetime(2024, 1, 1, 18) in d
    assert datetime.datetime(2024, 1, 2, 0) in d
    assert datetime.datetime(2024, 1, 2, 6) in d
    assert datetime.datetime(2024, 1, 2, 12) in d
    assert datetime.datetime(2024, 1, 2, 18) in d
    assert len(list(d)) == 10


def test_date_hindcast_1() -> None:
    """Test datetimes_factory with hindcast configuration."""
    d = _(
        """
        - name: hindcast
          reference_dates:
            start: 2023-01-01
            end: 2023-01-03
            frequency: 24
          years: 20
    """
    )
    assert len(list(d)) == 60


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
from collections.abc import Iterator


class HindcastDatesTimes:
    """The HindcastDatesTimes class is an iterator that generates datetime objects within a given range.

    Attributes
    ----------
    reference_dates : List[datetime.datetime]
        List of reference dates.
    years : int
        Number of years to go back from each reference date.
    """

    def __init__(self, reference_dates: list[datetime.datetime], years: int = 20):
        """Initialize the HindcastDatesTimes iterator.

        Parameters
        ----------
        reference_dates : List[datetime.datetime]
            List of reference dates.
        years : int, optional
            Number of years to go back from each reference date, by default 20.
        """
        self.reference_dates = reference_dates

        assert isinstance(years, int), f"years must be an integer, got {years}"
        assert years > 0, f"years must be greater than 0, got {years}"
        self.years = years

    def __iter__(self) -> Iterator[tuple[datetime.datetime, datetime.datetime]]:
        """Generate tuples of past dates and their corresponding reference dates.

        Yields
        ------
        Iterator[Tuple[datetime.datetime, datetime.datetime]]
            Tuples containing a past date and its corresponding reference date.
        """
        for reference_date in self.reference_dates:
            year, month, day = reference_date.year, reference_date.month, reference_date.day
            if (month, day) == (2, 29):
                day = 28

            for i in range(1, self.years + 1):
                date = datetime.datetime(year - i, month, day)
                yield (date, reference_date)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


"""Logging utilities."""

import logging
import threading

thread_local = threading.local()


LOGGER = logging.getLogger(__name__)


def set_logging_name(name: str) -> None:
    """Set the logging name for the current thread.

    Parameters
    ----------
    name : str
        The name to set for logging.
    """
    thread_local.logging_name = name


class ThreadCustomFormatter(logging.Formatter):
    """Custom logging formatter that includes thread-specific logging names."""

    def format(self, record: logging.LogRecord) -> str:
        """Format the log record to include the thread-specific logging name.

        Parameters
        ----------
        record : logging.LogRecord
            The log record to format.

        Returns
        -------
        str
            The formatted log record.
        """
        record.logging_name = thread_local.logging_name
        return super().format(record)


def enable_logging_name(name: str = "main") -> None:
    """Enable logging with a thread-specific logging name.

    Parameters
    ----------
    name : str, optional
        The default logging name to set, by default "main".
    """
    thread_local.logging_name = name

    formatter = ThreadCustomFormatter("%(asctime)s - %(logging_name)s - %(levelname)s - %(message)s")

    logger = logging.getLogger()

    for handler in logger.handlers:
        handler.setFormatter(formatter)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import calendar
import datetime
import re
from typing import Any

import aniso8601


def normalise_frequency(frequency: int | str) -> int:
    """Normalise frequency to hours.

    Parameters
    ----------
    frequency : int or str
        The frequency to normalise.

    Returns
    -------
    int
        The normalised frequency in hours.
    """
    if isinstance(frequency, int):
        return frequency
    assert isinstance(frequency, str), (type(frequency), frequency)

    unit = frequency[-1].lower()
    v = int(frequency[:-1])
    return {"h": v, "d": v * 24}[unit]


def _no_time_zone(date: datetime.datetime) -> datetime.datetime:
    """Remove time zone information from a date.

    Parameters
    ----------
    date : datetime.datetime
        A datetime object.

    Returns
    -------
    datetime.datetime
        The datetime object without time zone information.
    """

    return date.replace(tzinfo=None)


# this function is use in anemoi-datasets
def as_datetime(date: datetime.date | datetime.datetime | str, keep_time_zone: bool = False) -> datetime.datetime:
    """Convert a date to a datetime object, removing any time zone information.

    Parameters
    ----------
    date : datetime.date or datetime.datetime or str
        The date to convert.
    keep_time_zone : bool, optional
        If True, the time zone information is kept, by default False.

    Returns
    -------
    datetime.datetime
        The datetime object.
    """

    tidy = _no_time_zone if not keep_time_zone else lambda x: x

    if isinstance(date, datetime.datetime):
        return tidy(date)

    if isinstance(date, datetime.date):
        return tidy(datetime.datetime(date.year, date.month, date.day))

    if isinstance(date, str):
        return tidy(datetime.datetime.fromisoformat(date))

    raise ValueError(f"Invalid date type: {type(date)}")


def _as_datetime_list(date: datetime.date | datetime.datetime | str, default_increment: datetime.timedelta) -> iter:
    """Convert a date to a list of datetime objects.

    Parameters
    ----------
    date : datetime.date or datetime.datetime or str
        The date to convert.
    default_increment : datetime.timedelta
        The default increment for the list.

    Returns
    -------
    iter
        An iterator of datetime objects.
    """
    if isinstance(date, (list, tuple)):
        for d in date:
            yield from _as_datetime_list(d, default_increment)

    if isinstance(date, str):
        # Check for ISO format
        try:
            start, end = aniso8601.parse_interval(date)
            while start <= end:
                yield as_datetime(start)
                start += default_increment

            return

        except aniso8601.exceptions.ISOFormatError:
            pass

        try:
            intervals = aniso8601.parse_repeating_interval(date)
            for date in intervals:
                yield as_datetime(date)
            return
        except aniso8601.exceptions.ISOFormatError:
            pass

    yield as_datetime(date)


def as_datetime_list(
    date: datetime.date | datetime.datetime | str, default_increment: int = 1
) -> list[datetime.datetime]:
    """Convert a date to a list of datetime objects.

    Parameters
    ----------
    date : datetime.date or datetime.datetime or str
        The date to convert.
    default_increment : int, optional
        The default increment in hours, by default 1.

    Returns
    -------
    list of datetime.datetime
        A list of datetime objects.
    """
    default_increment = frequency_to_timedelta(default_increment)
    return list(_as_datetime_list(date, default_increment))


def as_timedelta(frequency: int | str | datetime.timedelta) -> datetime.timedelta:
    """Convert anything to a timedelta object.

    Parameters
    ----------
    frequency : int or str or datetime.timedelta
        The frequency to convert. If an integer, it is assumed to be in hours. If a string, it can be in the format:

        - "1h" for 1 hour
        - "1d" for 1 day
        - "1m" for 1 minute
        - "1s" for 1 second
        - "1:30" for 1 hour and 30 minutes
        - "1:30:10" for 1 hour, 30 minutes and 10 seconds
        - "PT10M" for 10 minutes (ISO8601)

        If a timedelta object is provided, it is returned as is.

    Returns
    -------
    datetime.timedelta
        The timedelta object.

    Raises
    ------
    ValueError
        Exception raised if the frequency cannot be converted to a timedelta.
    """

    if isinstance(frequency, datetime.timedelta):
        return frequency

    if isinstance(frequency, int):
        return datetime.timedelta(hours=frequency)

    assert isinstance(frequency, str), (type(frequency), frequency)

    try:
        return frequency_to_timedelta(int(frequency))
    except ValueError:
        pass

    if frequency.startswith(" ") or frequency.startswith(" "):
        frequency = frequency.strip()

    if frequency.startswith("-"):
        return -as_timedelta(frequency[1:])

    if frequency.startswith("+"):
        return as_timedelta(frequency[1:])

    if re.match(r"^\d+[hdms]$", frequency, re.IGNORECASE):
        unit = frequency[-1].lower()
        v = int(frequency[:-1])
        unit = {"h": "hours", "d": "days", "s": "seconds", "m": "minutes"}[unit]
        return datetime.timedelta(**{unit: v})

    if re.match(r"^\d+:\d+(:\d+)?$", frequency):
        m = frequency.split(":")
        if len(m) == 2:
            return datetime.timedelta(hours=int(m[0]), minutes=int(m[1]))

        if len(m) == 3:
            return datetime.timedelta(hours=int(m[0]), minutes=int(m[1]), seconds=int(m[2]))

    if re.match(r"^\d+ days?, \d+:\d+:\d+$", frequency):
        m = frequency.split(", ")
        days = int(m[0].split()[0])
        hms = m[1].split(":")
        return datetime.timedelta(days=days, hours=int(hms[0]), minutes=int(hms[1]), seconds=int(hms[2]))

    # ISO8601
    try:
        return aniso8601.parse_duration(frequency)
    except aniso8601.exceptions.ISOFormatError:
        pass

    raise ValueError(f"Cannot convert frequency {frequency} to timedelta")


def frequency_to_timedelta(frequency: int | str | datetime.timedelta) -> datetime.timedelta:
    """Convert a frequency to a timedelta object.

    Parameters
    ----------
    frequency : int or str or datetime.timedelta
        The frequency to convert.

    Returns
    -------
    datetime.timedelta
        The timedelta object.
    """
    return as_timedelta(frequency)


def frequency_to_string(frequency: datetime.timedelta) -> str:
    """Convert a frequency (i.e. a datetime.timedelta) to a string.

    Parameters
    ----------
    frequency : datetime.timedelta
        The frequency to convert.

    Returns
    -------
    str
        A string representation of the frequency.
    """

    frequency = frequency_to_timedelta(frequency)

    total_seconds = frequency.total_seconds()
    if total_seconds < 0:
        return f"-{frequency_to_string(-frequency)}"
    assert int(total_seconds) == total_seconds, total_seconds
    total_seconds = int(total_seconds)

    seconds = total_seconds

    days = seconds // (24 * 3600)
    seconds %= 24 * 3600
    hours = seconds // 3600
    seconds %= 3600
    minutes = seconds // 60
    seconds %= 60

    if days > 0 and hours == 0 and minutes == 0 and seconds == 0:
        return f"{days}d"

    if days == 0 and hours > 0 and minutes == 0 and seconds == 0:
        return f"{hours}h"

    if days == 0 and hours == 0 and minutes > 0 and seconds == 0:
        return f"{minutes}m"

    if days == 0 and hours == 0 and minutes == 0 and seconds > 0:
        return f"{seconds}s"

    if days > 0:
        return f"{total_seconds}s"

    return str(frequency)


def frequency_to_seconds(frequency: int | str | datetime.timedelta) -> int:
    """Convert a frequency to seconds.

    Parameters
    ----------
    frequency : int or str or datetime.timedelta
        The frequency to convert.

    Returns
    -------
    int
        Number of seconds.
    """
    result = frequency_to_timedelta(frequency).total_seconds()
    assert int(result) == result, result
    return int(result)


DOW = {
    "monday": 0,
    "tuesday": 1,
    "wednesday": 2,
    "thursday": 3,
    "friday": 4,
    "saturday": 5,
    "sunday": 6,
}


MONTH = {
    "january": 1,
    "february": 2,
    "march": 3,
    "april": 4,
    "may": 5,
    "june": 6,
    "july": 7,
    "august": 8,
    "september": 9,
    "october": 10,
    "november": 11,
    "december": 12,
}


def _make_day(day: tuple[int, list[int]] | None) -> set[int]:
    """Create a set of days.

    Parameters
    ----------
    day : int or list of int or None
        The day(s) to include in the set.

    Returns
    -------
    set of int
        A set of days.
    """
    if day is None:
        return set(range(1, 32))
    if not isinstance(day, list):
        day = [day]
    return {int(d) for d in day}


def _make_week(week: tuple[str, list[str]] | None) -> set[int]:
    """Create a set of weekdays.

    Parameters
    ----------
    week : str or list of str or None
        The weekday(s) to include in the set.

    Returns
    -------
    set of int
        A set of weekdays.
    """
    if week is None:
        return set(range(7))
    if not isinstance(week, list):
        week = [week]
    return {DOW[w.lower()] for w in week}


def _make_months(months: int | str | list[int | str] | None) -> set[int]:
    """Create a set of months.

    Parameters
    ----------
    months : int or str or list of int or str or None
        The month(s) to include in the set.

    Returns
    -------
    set of int
        A set of months.
    """
    if months is None:
        return set(range(1, 13))

    if not isinstance(months, list):
        months = [months]

    return {int(MONTH.get(m, m)) for m in months}


class DateTimes:
    """The DateTimes class is an iterator that generates datetime objects within a given range."""

    def __init__(
        self,
        start: datetime.date | datetime.datetime | str,
        end: datetime.date | datetime.datetime | str,
        increment: int = 24,
        *,
        day_of_month: tuple[int, list[int]] | None = None,
        day_of_week: tuple[str, list[str]] | None = None,
        calendar_months: int | str | list[int | str] | None = None,
    ):
        """Initialize the DateTimes iterator.

        Parameters
        ----------
        start : datetime.date or datetime.datetime or str
            The start date.
        end : datetime.date or datetime.datetime or str
            The end date.
        increment : int, optional
            The increment in hours, by default 24.
        day_of_month : int or list of int or None, optional
            The day(s) of the month to include, by default None.
        day_of_week : str or list of str or None, optional
            The day(s) of the week to include, by default None.
        calendar_months : int or str or list of int or str or None, optional
            The month(s) to include, by default None.
        """
        self.start = as_datetime(start)
        self.end = as_datetime(end)
        self.increment = frequency_to_timedelta(increment)
        self.day_of_month = _make_day(day_of_month)
        self.day_of_week = _make_week(day_of_week)
        self.calendar_months = _make_months(calendar_months)

    def __iter__(self) -> iter:
        """Iterate over the datetime objects.

        Returns
        -------
        iter
            An iterator of datetime objects.
        """
        date = self.start
        while date <= self.end:
            if (
                (date.weekday() in self.day_of_week)
                and (date.day in self.day_of_month)
                and (date.month in self.calendar_months)
            ):

                yield date
            date += self.increment


class Year(DateTimes):
    """Year is defined as the months of January to December."""

    def __init__(self, year: int, **kwargs):
        """Initialize the Year iterator.

        Parameters
        ----------
        year : int
            The year.
        """
        super().__init__(datetime.datetime(year, 1, 1), datetime.datetime(year, 12, 31), **kwargs)


class Winter(DateTimes):
    """Winter is defined as the months of December, January and February."""

    def __init__(self, year: int, **kwargs):
        """Initialize the Winter iterator.

        Parameters
        ----------
        year : int
            The year.
        """
        super().__init__(
            datetime.datetime(year, 12, 1),
            datetime.datetime(year + 1, 2, calendar.monthrange(year + 1, 2)[1]),
            **kwargs,
        )


class Spring(DateTimes):
    """Spring is defined as the months of March, April and May."""

    def __init__(self, year: int, **kwargs):
        """Initialize the Spring iterator.

        Parameters
        ----------
        year : int
            The year.
        """
        super().__init__(datetime.datetime(year, 3, 1), datetime.datetime(year, 5, 31), **kwargs)


class Summer(DateTimes):
    """Summer is defined as the months of June, July and August."""

    def __init__(self, year: int, **kwargs):
        """Initialize the Summer iterator.

        Parameters
        ----------
        year : int
            The year.
        """
        super().__init__(datetime.datetime(year, 6, 1), datetime.datetime(year, 8, 31), **kwargs)


class Autumn(DateTimes):
    """Autumn is defined as the months of September, October and November."""

    def __init__(self, year: int, **kwargs):
        """Initialize the Autumn iterator.

        Parameters
        ----------
        year : int
            The year.
        """
        super().__init__(datetime.datetime(year, 9, 1), datetime.datetime(year, 11, 30), **kwargs)


class ConcatDateTimes:
    """ConcatDateTimes is an iterator that generates datetime objects from a list of dates."""

    def __init__(self, *dates: DateTimes):
        """Initialize the ConcatDateTimes iterator.

        Parameters
        ----------
        dates : DateTimes
            The list of DateTimes objects.
        """
        if len(dates) == 1 and isinstance(dates[0], list):
            dates = dates[0]

        self.dates = dates

    def __iter__(self) -> iter:
        """Iterate over the datetime objects.

        Returns
        -------
        iter
            An iterator of datetime objects.
        """
        for date in self.dates:
            yield from date


class EnumDateTimes:
    """EnumDateTimes is an iterator that generates datetime objects from a list of dates."""

    def __init__(self, dates: list[datetime.date | datetime.datetime | str]):
        """Initialize the EnumDateTimes iterator.

        Parameters
        ----------
        dates : list of datetime.date or datetime.datetime or str
            The list of dates.
        """
        self.dates = dates

    def __iter__(self) -> iter:
        """Iterate over the datetime objects.

        Returns
        -------
        iter
            An iterator of datetime objects.
        """
        for date in self.dates:
            yield as_datetime(date)


def datetimes_factory(*args: Any, **kwargs: Any) -> DateTimes | ConcatDateTimes | EnumDateTimes:
    """Create a DateTimes, ConcatDateTimes, or EnumDateTimes object.

    Parameters
    ----------
    *args : Any
        Positional arguments.
    **kwargs : Any
        Keyword arguments.

    Returns
    -------
    DateTimes or ConcatDateTimes or EnumDateTimes
        The created object.
    """
    if args and kwargs:
        raise ValueError("Cannot provide both args and kwargs for a list of dates")

    if not args and not kwargs:
        raise ValueError("No dates provided")

    if kwargs:
        name = kwargs.get("name")

        if name == "hindcast":
            from .hindcasts import HindcastDatesTimes

            reference_dates = kwargs["reference_dates"]
            reference_dates = datetimes_factory(reference_dates)
            years = kwargs["years"]
            return HindcastDatesTimes(reference_dates=reference_dates, years=years)

        kwargs = kwargs.copy()
        if "frequency" in kwargs:
            kwargs["increment"] = kwargs.pop("frequency")
        return DateTimes(**kwargs)

    if not any((isinstance(x, dict) or isinstance(x, list)) for x in args):
        return EnumDateTimes(args)

    if len(args) == 1:
        a = args[0]

        if isinstance(a, dict):
            return datetimes_factory(**a)

        if isinstance(a, list):
            return datetimes_factory(*a)

    return ConcatDateTimes(*[datetimes_factory(a) for a in args])


if __name__ == "__main__":
    print(as_datetime_list("R10/2023-01-01T00:00:00Z/P1D"))
    print(as_datetime_list("2007-03-01T13:00:00/2008-05-11T15:30:00", "200h"))



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


"""Read and write extra metadata in PyTorch checkpoints files. These files
are zip archives containing the model weights.
"""

import json
import logging
import os
import time
import zipfile
from collections.abc import Callable
from tempfile import TemporaryDirectory

import tqdm

LOG = logging.getLogger(__name__)

DEFAULT_NAME = "ai-models.json"
DEFAULT_FOLDER = "anemoi-metadata"


def has_metadata(path: str, *, name: str = DEFAULT_NAME) -> bool:
    """Check if a checkpoint file has a metadata file.

    Parameters
    ----------
    path : str
        The path to the checkpoint file
    name : str, optional
        The name of the metadata file in the zip archive

    Returns
    -------
    bool
        True if the metadata file is found
    """
    with zipfile.ZipFile(path, "r") as f:
        for b in f.namelist():
            if os.path.basename(b) == name:
                return True
    return False


def metadata_root(path: str, *, name: str = DEFAULT_NAME) -> str:
    """Get the root directory of the metadata file.

    Parameters
    ----------
    path : str
        The path to the checkpoint file
    name : str, optional
        The name of the metadata file in the zip archive

    Returns
    -------
    str
        The root directory of the metadata file

    Raises
    ------
    ValueError
        If the metadata file is not found
    """
    with zipfile.ZipFile(path, "r") as f:
        for b in f.namelist():
            if os.path.basename(b) == name:
                return os.path.dirname(b)
    raise ValueError(f"Could not find '{name}' in {path}.")


def load_metadata(path: str, *, supporting_arrays: bool = False, name: str = DEFAULT_NAME) -> dict:
    """Load metadata from a checkpoint file.

    Parameters
    ----------
    path : str
        The path to the checkpoint file

    supporting_arrays : bool, optional
        If True, the function will return a dictionary with the supporting arrays

    name : str, optional
        The name of the metadata file in the zip archive

    Returns
    -------
    dict
        The content of the metadata file from JSON

    Raises
    ------
    ValueError
        If the metadata file is not found
    """
    with zipfile.ZipFile(path, "r") as f:
        metadata = None
        for b in f.namelist():
            if os.path.basename(b) == name:
                if metadata is not None:
                    raise ValueError(f"Found two or more '{name}' in {path}.")
                metadata = b

    if metadata is not None:
        with zipfile.ZipFile(path, "r") as f:
            metadata = json.load(f.open(metadata, "r"))
            if supporting_arrays:
                arrays = load_supporting_arrays(f, metadata.get("supporting_arrays_paths", {}))
                return metadata, arrays

            return metadata
    else:
        raise ValueError(f"Could not find '{name}' in {path}.")


def load_supporting_arrays(zipf: zipfile.ZipFile, entries: dict) -> dict:
    """Load supporting arrays from a zip file.

    Parameters
    ----------
    zipf : zipfile.ZipFile
        The zip file
    entries : dict
        A dictionary of entries with paths, shapes, and dtypes

    Returns
    -------
    dict
        A dictionary of supporting arrays
    """
    import numpy as np

    supporting_arrays = {}
    for key, entry in entries.items():
        supporting_arrays[key] = np.frombuffer(
            zipf.read(entry["path"]),
            dtype=entry["dtype"],
        ).reshape(entry["shape"])
    return supporting_arrays


def save_metadata(
    path: str, metadata: dict, *, supporting_arrays: dict = None, name: str = DEFAULT_NAME, folder: str = DEFAULT_FOLDER
) -> None:
    """Save metadata to a checkpoint file.

    Parameters
    ----------
    path : str
        The path to the checkpoint file
    metadata : dict
        A JSON serializable object
    supporting_arrays : dict, optional
        A dictionary of supporting NumPy arrays
    name : str, optional
        The name of the metadata file in the zip archive
    folder : str, optional
        The folder where the metadata file will be saved
    """
    with zipfile.ZipFile(path, "a") as zipf:

        directories = set()

        for b in zipf.namelist():
            directory = os.path.dirname(b)
            while os.path.dirname(directory) not in (".", ""):
                directory = os.path.dirname(directory)
            directories.add(directory)

            if os.path.basename(b) == name:
                raise ValueError(f"'{name}' already in {path}")

        if len(directories) != 1:
            # PyTorch checkpoints should have a single directory
            # otherwise PyTorch will complain
            raise ValueError(f"No or multiple directories in the checkpoint {path}, directories={directories}")

        directory = list(directories)[0]

        LOG.info("Adding extra information to checkpoint %s", path)
        LOG.info("Saving metadata to %s/%s/%s", directory, folder, name)

        metadata = metadata.copy()
        if supporting_arrays is not None:
            metadata["supporting_arrays_paths"] = {
                key: dict(path=f"{directory}/{folder}/{key}.numpy", shape=value.shape, dtype=str(value.dtype))
                for key, value in supporting_arrays.items()
            }
        else:
            metadata["supporting_arrays_paths"] = {}

        zipf.writestr(
            f"{directory}/{folder}/{name}",
            json.dumps(metadata),
        )

        for name, entry in metadata["supporting_arrays_paths"].items():
            value = supporting_arrays[name]
            LOG.info(
                "Saving supporting array `%s` to %s (shape=%s, dtype=%s)",
                name,
                entry["path"],
                entry["shape"],
                entry["dtype"],
            )
            zipf.writestr(entry["path"], value.tobytes())


def _edit_metadata(path: str, name: str, callback: Callable, supporting_arrays: dict = None) -> None:
    """Edit metadata in a checkpoint file.

    Parameters
    ----------
    path : str
        The path to the checkpoint file
    name : str
        The name of the metadata file in the zip archive
    callback : Callable
        A callback function to edit the metadata
    supporting_arrays : dict, optional
        A dictionary of supporting NumPy arrays
    """
    new_path = f"{path}.anemoi-edit-{time.time()}-{os.getpid()}.tmp"

    found = False

    directory = None
    with TemporaryDirectory() as temp_dir:
        zipfile.ZipFile(path, "r").extractall(temp_dir)
        total = 0
        for root, dirs, files in os.walk(temp_dir):
            for f in files:
                total += 1
                full = os.path.join(root, f)
                if f == name:
                    found = True
                    callback(full)
                    directory = os.path.dirname(full)

        if not found:
            raise ValueError(f"Could not find '{name}' in {path}")

        if supporting_arrays is not None:

            for key, entry in supporting_arrays.items():
                value = entry.tobytes()
                fname = os.path.join(directory, f"{key}.numpy")
                os.makedirs(os.path.dirname(fname), exist_ok=True)
                with open(fname, "wb") as f:
                    f.write(value)
                    total += 1

        with zipfile.ZipFile(new_path, "w", zipfile.ZIP_DEFLATED) as zipf:
            with tqdm.tqdm(total=total, desc="Rebuilding checkpoint") as pbar:
                for root, dirs, files in os.walk(temp_dir):
                    for f in files:
                        full = os.path.join(root, f)
                        rel = os.path.relpath(full, temp_dir)
                        zipf.write(full, rel)
                        pbar.update(1)

    os.rename(new_path, path)
    LOG.info("Updated metadata in %s", path)


def replace_metadata(path: str, metadata: dict, supporting_arrays: dict = None, *, name: str = DEFAULT_NAME) -> None:
    """Replace metadata in a checkpoint file.

    Parameters
    ----------
    path : str
        The path to the checkpoint file
    metadata : dict
        A JSON serializable object
    supporting_arrays : dict, optional
        A dictionary of supporting NumPy arrays
    name : str, optional
        The name of the metadata file in the zip archive
    """
    if not isinstance(metadata, dict):
        raise ValueError(f"metadata must be a dict, got {type(metadata)}")

    if "version" not in metadata:
        raise ValueError("metadata must have a 'version' key")

    def callback(full):
        with open(full, "w") as f:
            json.dump(metadata, f)

    return _edit_metadata(path, name, callback, supporting_arrays)


def remove_metadata(path: str, *, name: str = DEFAULT_NAME) -> None:
    """Remove metadata from a checkpoint file.

    Parameters
    ----------
    path : str
        The path to the checkpoint file
    name : str, optional
        The name of the metadata file in the zip archive
    """

    def callback(full):
        os.remove(full)

    return _edit_metadata(path, name, callback)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


"""Logging utilities."""

import logging
import time
from collections import defaultdict
from typing import Any

from .humanize import seconds_to_human

LOGGER = logging.getLogger(__name__)


class Timer:
    """Context manager to measure elapsed time.

    Parameters
    ----------
    title : str
        The title of the timer.
    logger : logging.Logger, optional
        The logger to use for logging the elapsed time, by default LOGGER.
    """

    def __init__(self, title: str, logger: logging.Logger = LOGGER):
        """Initialize the Timer.

        Parameters
        ----------
        title : str
            The title of the timer.
        logger : logging.Logger, optional
            The logger to use for logging the elapsed time, by default LOGGER.
        """
        self.title = title
        self.start = time.time()
        self.logger = logger

    def __enter__(self) -> "Timer":
        """Enter the runtime context related to this object.

        Returns
        -------
        Timer
            The Timer object itself.
        """
        return self

    @property
    def elapsed(self) -> float:
        """Float: The elapsed time in seconds."""
        return time.time() - self.start

    def __exit__(self, *args: Any) -> None:
        """Exit the runtime context related to this object.

        Parameters
        ----------
        *args : Any
            Exception information.
        """
        self.logger.info("%s: %s.", self.title, seconds_to_human(self.elapsed))


class _Timer:
    """Internal timer class."""

    def __init__(self):
        """Initialize the _Timer."""
        self.elapsed = 0.0

    def __enter__(self) -> "_Timer":
        """Enter the runtime context related to this object.

        Returns
        -------
        _Timer
            The _Timer object itself.
        """
        self.start()
        return self

    def __exit__(self, *args: Any) -> None:
        """Exit the runtime context related to this object.

        Parameters
        ----------
        *args : Any
            Exception information.
        """
        self.stop()

    def start(self) -> None:
        """Start the timer."""
        self._start = time.time()

    def stop(self) -> None:
        """Stop the timer and accumulate the elapsed time."""
        self.elapsed += time.time() - self._start


class Timers:
    """A collection of timers.

    Parameters
    ----------
    logger : logging.Logger, optional
        The logger to use for logging the elapsed time, by default LOGGER.
    """

    def __init__(self, logger: logging.Logger = LOGGER):
        """Initialize the Timers collection.

        Parameters
        ----------
        logger : logging.Logger, optional
            The logger to use for logging the elapsed time, by default LOGGER.
        """
        self.logger = logger
        self.timers = defaultdict(_Timer)

    def __getitem__(self, name: str) -> _Timer:
        """Get a timer by name.

        Parameters
        ----------
        name : str
            The name of the timer.

        Returns
        -------
        _Timer
            The timer with the given name.
        """
        return self.timers[name]

    def report(self) -> None:
        """Log the elapsed time for all timers."""
        length = max(len(name) for name in self.timers)
        for name, timer in sorted(self.timers.items()):
            self.logger.info("%s: %s.", f"{name:<{length}}", seconds_to_human(timer.elapsed))



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


"""Utilities for working with Mars requests.

Has some knowledge of how certain streams are organised in Mars.
"""

import datetime
import logging
import os
from typing import Any

import yaml

LOG = logging.getLogger(__name__)

DEFAULT_MARS_LABELLING = {
    "class": "od",
    "type": "an",
    "stream": "oper",
    "expver": "0001",
}


def _expand_mars_labelling(request: dict[str, Any]) -> dict[str, Any]:
    """Expand the request with the default Mars labelling.

    Parameters
    ----------
    request : dict
        The original Mars request.

    Returns
    -------
    dict
        The Mars request expanded with default labelling.
    """
    result = DEFAULT_MARS_LABELLING.copy()
    result.update(request)
    return result


STREAMS = None


def _lookup_mars_stream(request: dict[str, Any]) -> dict[str, Any] | None:
    """Look up the Mars stream information for a given request.

    Parameters
    ----------
    request : dict
        The Mars request.

    Returns
    -------
    dict or None
        The stream information if a match is found, otherwise None.
    """
    global STREAMS

    if STREAMS is None:

        with open(os.path.join(os.path.dirname(__file__), "mars.yaml")) as f:
            STREAMS = yaml.safe_load(f)

    request = _expand_mars_labelling(request)
    for s in STREAMS:
        match = s["match"]
        if all(request.get(k) == v for k, v in match.items()):
            return s["info"]


def recenter(
    date: datetime.datetime, center: dict[str, Any], members: dict[str, Any]
) -> tuple[dict[str, Any] | None, dict[str, Any] | None]:
    """Recenter the given date with the specified center and members.

    Parameters
    ----------
    date : datetime.datetime
        The date to recenter.
    center : dict
        The center request information.
    members : dict
        The members request information.

    Returns
    -------
    tuple
        A tuple containing the recentered center and members information.
    """
    center = _lookup_mars_stream(center)
    members = _lookup_mars_stream(members)

    return (center, members)


if __name__ == "__main__":
    date = datetime.datetime(2024, 5, 9, 0)

    print(recenter(date, {"type": "an"}, {"stream": "elda"}))



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
import random
import shlex
import subprocess

from ..humanize import bytes_to_human
from . import BaseUpload

LOGGER = logging.getLogger(__name__)


def call_process(*args: str) -> str:
    """Execute a subprocess with the given arguments and return its output.

    Parameters
    ----------
    args : str
        The command and its arguments to execute.

    Returns
    -------
    str
        The standard output of the command.

    Raises
    ------
    RuntimeError
        If the command returns a non-zero exit code.
    """
    proc = subprocess.Popen(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = proc.communicate()
    if proc.returncode != 0:
        print(stdout)
        msg = f"{' '.join(args)} failed: {stderr}"
        raise RuntimeError(msg)

    return stdout.decode("utf-8").strip()


class SshBaseUpload(BaseUpload):

    def _parse_target(self, target: str) -> tuple[str, str]:
        """Parse the SSH target string into hostname and path.

        Parameters
        ----------
        target : str
            The SSH target string in the format 'ssh://hostname:path'.

        Returns
        -------
        tuple[str, str]
            A tuple containing the hostname and the path.

        Raises
        ------
        Exception
            If the path contains suspicious '..'.
        """
        assert target.startswith("ssh://"), target

        target = target[6:]
        hostname, path = target.split(":")

        if "+" in hostname:
            hostnames = hostname.split("+")
            hostname = hostnames[random.randint(0, len(hostnames) - 1)]

        if ".." in path.split("/"):
            raise Exception("Path contains suspicious '..' : {target}")

        return hostname, path

    def get_temporary_target(self, target: str, pattern: str) -> str:
        """Get a temporary target path based on the given pattern.

        Parameters
        ----------
        target : str
            The original target path.
        pattern : str
            The pattern to format the temporary path.

        Returns
        -------
        str
            The temporary target path.
        """
        hostname, path = self._parse_target(target)
        if pattern is not None:
            dirname, basename = os.path.split(path)
            path = pattern.format(dirname=dirname, basename=basename)
        return f"ssh://{hostname}:{path}"

    def rename_target(self, target: str, new_target: str) -> None:
        """Rename the target to a new target path.

        Parameters
        ----------
        target : str
            The original target path.
        new_target : str
            The new target path.
        """
        hostname, path = self._parse_target(target)
        hostname, new_path = self._parse_target(new_target)
        call_process("ssh", hostname, "mkdir", "-p", shlex.quote(os.path.dirname(new_path)))
        call_process("ssh", hostname, "mv", shlex.quote(path), shlex.quote(new_path))

    def delete_target(self, target: str) -> None:
        """Delete the target path.

        Parameters
        ----------
        target : str
            The target path to delete.
        """
        pass
        # hostname, path = self._parse_target(target)
        # LOGGER.info(f"Deleting {target}")
        # call_process("ssh", hostname, "rm", "-rf", shlex.quote(path))


class RsyncUpload(SshBaseUpload):

    def _transfer_file(
        self, source: str, target: str, overwrite: bool, resume: bool, verbosity: int, threads: int, config: dict = None
    ) -> int:
        """Transfer a file using rsync.

        Parameters
        ----------
        source : str
            The source file path.
        target : str
            The target file path.
        overwrite : bool
            Whether to overwrite the target if it exists.
        resume : bool
            Whether to resume the transfer if possible.
        verbosity : int
            The verbosity level.
        threads : int
            The number of threads to use.
        config : dict, optional
            Additional configuration options.

        Returns
        -------
        int
            The size of the transferred file.
        """
        hostname, path = self._parse_target(target)

        size = os.path.getsize(source)

        if verbosity > 0:
            LOGGER.info(f"{self.action} {source} to {target} ({bytes_to_human(size)})")

        call_process("ssh", hostname, "mkdir", "-p", shlex.quote(os.path.dirname(path)))
        call_process(
            "rsync",
            "-av",
            "--partial",
            # it would be nice to avoid two ssh calls, but the following is not possible,
            # this is because it requires a shell command and would not be safe.
            # # f"--rsync-path='mkdir -p {os.path.dirname(path)} && rsync'",
            source,
            f"{hostname}:{path}",
        )
        return size


class ScpUpload(SshBaseUpload):

    def _transfer_file(
        self, source: str, target: str, overwrite: bool, resume: bool, verbosity: int, threads: int, config: dict = None
    ) -> int:
        """Transfer a file using scp.

        Parameters
        ----------
        source : str
            The source file path.
        target : str
            The target file path.
        overwrite : bool
            Whether to overwrite the target if it exists.
        resume : bool
            Whether to resume the transfer if possible.
        verbosity : int
            The verbosity level.
        threads : int
            The number of threads to use.
        config : dict, optional
            Additional configuration options.

        Returns
        -------
        int
            The size of the transferred file.

        Raises
        ------
        ValueError
            If the target already exists and overwrite or resume is not specified.
        """
        hostname, path = self._parse_target(target)

        size = os.path.getsize(source)

        if verbosity > 0:
            LOGGER.info(f"{self.action} {source} to {target} ({bytes_to_human(size)})")

        remote_size = None
        try:
            out = call_process("ssh", hostname, "stat", "-c", "%s", shlex.quote(path))
            remote_size = int(out)
        except RuntimeError:
            remote_size = None

        if remote_size is not None:
            if remote_size != size:
                LOGGER.warning(
                    f"{target} already exists, but with different size, re-uploading (remote={remote_size}, local={size})"
                )
            elif resume:
                # LOGGER.info(f"{target} already exists, skipping")
                return size

        if remote_size is not None and not overwrite and not resume:
            raise ValueError(f"{target} already exists, use 'overwrite' to replace or 'resume' to skip")

        call_process("ssh", hostname, "mkdir", "-p", shlex.quote(os.path.dirname(path)))
        call_process("scp", source, shlex.quote(f"{hostname}:{path}"))

        return size


def upload(source: str, target: str, **kwargs) -> None:
    """Upload a file or folder to the target location using rsync.

    Parameters
    ----------
    source : str
        The source file or folder path.
    target : str
        The target path.
    kwargs : dict
        Additional arguments for the transfer.
    """
    uploader = RsyncUpload()

    if os.path.isdir(source):
        uploader.transfer_folder(source=source, target=target, **kwargs)
    else:
        uploader.transfer_file(source=source, target=target, **kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


"""Generate human readable strings."""

import datetime
import json
import re
import warnings
from collections import defaultdict
from collections.abc import Callable
from collections.abc import Generator
from typing import Any

from anemoi.utils.dates import as_datetime


def bytes_to_human(n: float) -> str:
    """Convert a number of bytes to a human readable string.

    >>> bytes_to_human(4096)
    '4 KiB'

    >>> bytes_to_human(4000)
    '3.9 KiB'

    Parameters
    ----------
    n : float
        The number of bytes

    Returns
    -------
    str
        A human readable string
    """
    if n < 0:
        sign = "-"
        n = -n
    else:
        sign = ""

    u = ["", " KiB", " MiB", " GiB", " TiB", " PiB", " EiB", " ZiB", " YiB"]
    i = 0
    while n >= 1024:
        n /= 1024.0
        i += 1
    return f"{sign}{int(n * 10 + 0.5) / 10.0:g}{u[i]}"


def bytes(n: float) -> str:
    """Deprecated function to convert bytes to a human readable string.

    Parameters
    ----------
    n : float
        The number of bytes

    Returns
    -------
    str
        A human readable string
    """
    warnings.warn(
        "Function bytes is deprecated and will be removed in a future version. Use bytes_to_human instead.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    return bytes_to_human(n)


def base2_to_human(n: float) -> str:
    """Convert a number to a human readable string using base 2 units.

    Parameters
    ----------
    n : float
        The number to convert

    Returns
    -------
    str
        A human readable string
    """
    u = ["", "K", "M", "G", "T", " P", "E", "Z", "Y"]
    i = 0
    while n >= 1024:
        n /= 1024.0
        i += 1
    return f"{int(n * 10 + 0.5) / 10.0:g}{u[i]}"


def base2(n: float) -> str:
    """Deprecated function to convert a number to a human readable string using base 2 units.

    Parameters
    ----------
    n : float
        The number to convert

    Returns
    -------
    str
        A human readable string
    """
    warnings.warn(
        "Function base2 is deprecated and will be removed in a future version. Use base2_to_human instead.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    return base2_to_human(n)


PERIODS = (
    (7 * 24 * 60 * 60, "week"),
    (24 * 60 * 60, "day"),
    (60 * 60, "hour"),
    (60, "minute"),
    (1, "second"),
)


def _plural(count: int) -> str:
    """Return 's' if count is not 1, otherwise return an empty string.

    Parameters
    ----------
    count : int
        The count

    Returns
    -------
    str
        's' if count is not 1, otherwise an empty string
    """
    if count != 1:
        return "s"
    else:
        return ""


def seconds_to_human(seconds: float | datetime.timedelta) -> str:
    """Convert a number of seconds to a human readable string.

    >>> seconds_to_human(4000)
    '1 hour 6 minutes 40 seconds'

    Parameters
    ----------
    seconds : float
        The number of seconds

    Returns
    -------
    str
        A human readable string
    """
    if isinstance(seconds, datetime.timedelta):
        seconds = seconds.total_seconds()

    if seconds == 0:
        return "instantaneous"

    if seconds < 0.1:
        units = [
            None,
            "milli",
            "micro",
            "nano",
            "pico",
            "femto",
            "atto",
            "zepto",
            "yocto",
        ]
        i = 0
        while seconds < 1.0 and i < len(units) - 1:
            seconds *= 1000
            i += 1
        if seconds > 100 and i > 0:
            seconds /= 1000
            i -= 1
        seconds = round(seconds * 10) / 10
        return f"{seconds:g} {units[i]}second{_plural(seconds)}"

    n = seconds
    s = []
    for p in PERIODS:
        m = int(n / p[0])
        if m:
            s.append("%d %s%s" % (m, p[1], _plural(m)))
            n %= p[0]

    if not s:
        seconds = round(seconds * 10) / 10
        s.append(f"{seconds:g} second{_plural(seconds)}")
    return " ".join(s)


def seconds(seconds: float) -> str:
    """Deprecated function to convert seconds to a human readable string.

    Parameters
    ----------
    seconds : float
        The number of seconds

    Returns
    -------
    str
        A human readable string
    """
    warnings.warn(
        "Function seconds is deprecated and will be removed in a future version. Use seconds_to_human instead.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    return seconds_to_human(seconds)


def plural(value: int, what: str) -> str:
    """Return a string with the value and the pluralized form of what.

    Parameters
    ----------
    value : int
        The value
    what : str
        The string to pluralize

    Returns
    -------
    str
        The value and the pluralized form of what
    """
    return f"{value:,} {what}{_plural(value)}"


DOW = [
    "Monday",
    "Tuesday",
    "Wednesday",
    "Thursday",
    "Friday",
    "Saturday",
    "Sunday",
]

MONTH = [
    "January",
    "February",
    "March",
    "April",
    "May",
    "June",
    "July",
    "August",
    "September",
    "October",
    "November",
    "December",
]


def __(n: int) -> str:
    """Return the ordinal suffix for a number.

    Parameters
    ----------
    n : int
        The number

    Returns
    -------
    str
        The ordinal suffix
    """
    if n in (11, 12, 13):
        return "th"

    if n % 10 == 1:
        return "st"

    if n % 10 == 2:
        return "nd"

    if n % 10 == 3:
        return "rd"

    return "th"


def when(
    then: datetime.datetime, now: datetime.datetime | None = None, short: bool = True, use_utc: bool = False
) -> str:
    """Generate a human readable string for a date, relative to now.

    >>> when(datetime.datetime.now() - datetime.timedelta(hours=2))
    '2 hours ago'

    >>> when(datetime.datetime.now() - datetime.timedelta(days=1))
    'yesterday at 08:46'

    >>> when(datetime.datetime.now() - datetime.timedelta(days=5))
    'last Sunday'

    >>> when(datetime.datetime.now() - datetime.timedelta(days=365))
    'last year'

    >>> when(datetime.datetime.now() + datetime.timedelta(days=365))
    'next year'

    Parameters
    ----------
    then : datetime.datetime
        A datetime
    now : datetime.datetime, optional
        The reference date, by default NOW
    short : bool, optional
        Generate shorter strings, by default True
    use_utc : bool, optional
        Use UTC time, by default False

    Returns
    -------
    str
        A human readable string
    """
    last = "last"

    if now is None:
        if use_utc:
            now = datetime.datetime.utcnow()
        else:
            now = datetime.datetime.now()

    diff = (now - then).total_seconds()

    if diff < 0:
        last = "next"
        diff = -diff

    diff = int(diff)

    if diff == 0:
        return "right now"

    def _(x):
        if last == "last":
            return f"{x} ago"
        else:
            return f"in {x}"

    if diff < 60:
        diff = int(diff + 0.5)
        return _(f"{diff} second{_plural(diff)}")

    if diff < 60 * 60:
        diff /= 60
        diff = int(diff + 0.5)
        return _(f"{diff} minute{_plural(diff)}")

    if diff < 60 * 60 * 6:
        diff /= 60 * 60
        diff = int(diff + 0.5)
        return _(f"{diff} hour{_plural(diff)}")

    jnow = now.toordinal()
    jthen = then.toordinal()

    if jnow == jthen:
        return "today at %02d:%02d" % (then.hour, then.minute)

    if jnow == jthen + 1:
        return "yesterday at %02d:%02d" % (then.hour, then.minute)

    if jnow == jthen - 1:
        return "tomorrow at %02d:%02d" % (then.hour, then.minute)

    if abs(jnow - jthen) <= 7:
        if last == "next":
            last = "this"
        return "{} {}".format(
            last,
            DOW[then.weekday()],
        )

    if abs(jnow - jthen) < 32 and now.month == then.month:
        return "the %d%s of this month" % (then.day, __(then.day))

    if abs(jnow - jthen) < 64 and now.month == then.month + 1:
        return "the %d%s of %s month" % (then.day, __(then.day), last)

    if short:
        years = int(abs(jnow - jthen) / 365.25 + 0.5)
        if years == 1:
            return "%s year" % last

        if years > 1:
            return _("%d years" % (years,))

        delta = abs(now - then)
        if delta.days > 1 and delta.days < 30:
            return _("%d days" % (delta.days,))

    return "on %s %d %s %d" % (
        DOW[then.weekday()],
        then.day,
        MONTH[then.month],
        then.year,
    )


def string_distance(s: str, t: str) -> int:
    """Calculate the Levenshtein distance between two strings.

    Parameters
    ----------
    s : str
        The first string
    t : str
        The second string

    Returns
    -------
    int
        The Levenshtein distance
    """
    import numpy as np

    m = len(s)
    n = len(t)
    d = np.zeros((m + 1, n + 1), dtype=int)

    one = int(1)
    zero = int(0)

    d[:, 0] = np.arange(m + 1)
    d[0, :] = np.arange(n + 1)

    for i in range(1, m + 1):
        for j in range(1, n + 1):
            cost = zero if s[i - 1] == t[j - 1] else one
            d[i, j] = min(
                d[i - 1, j] + one,
                d[i, j - 1] + one,
                d[i - 1, j - 1] + cost,
            )

    return d[m, n]


def did_you_mean(word: str, vocabulary: list[str]) -> str:
    """Pick the closest word in a vocabulary.

    >>> did_you_mean("aple", ["banana", "lemon", "apple", "orange"])
    'apple'

    Parameters
    ----------
    word : str
        The word to look for
    vocabulary : list of str
        The list of known words

    Returns
    -------
    str
        The closest word in the vocabulary
    """
    _, best = min((string_distance(word, w), w) for w in vocabulary)
    # if distance < min(len(word), len(best)):
    return best


def dict_to_human(query: dict[str, Any]) -> str:
    """Convert a dictionary to a human readable string.

    Parameters
    ----------
    query : dict
        The dictionary to convert

    Returns
    -------
    str
        A human readable string
    """
    lst = [f"{k}={v}" for k, v in sorted(query.items())]

    return list_to_human(lst)


def list_to_human(lst: list[str], conjunction: str = "and") -> str:
    """Convert a list of strings to a human readable string.

    >>> list_to_human(["banana", "lemon", "apple", "orange"])
    'banana, lemon, apple and orange'

    Parameters
    ----------
    lst : list of str
        The list of strings to concatenate
    conjunction : str, optional
        The word to connect the last word in the list (like "or" or "and"), by default "and"

    Returns
    -------
    str
        Human readable string of list
    """
    if not lst:
        return "??"

    if len(lst) > 2:
        lst = [", ".join(lst[:-1]), lst[-1]]

    return f" {conjunction} ".join(lst)


def human_to_number(value: str | int, name: str, units: dict[str, int], none_ok: bool) -> int | None:
    """Convert a human readable string to a number.

    Parameters
    ----------
    value : str or int
        The value to convert
    name : str
        The name of the value
    units : dict
        The units to use for conversion
    none_ok : bool
        Whether None is an acceptable value

    Returns
    -------
    int or None
        The converted value
    """
    if value is None and none_ok:
        return None

    value = str(value)
    # TODO: support floats
    m = re.search(r"^\s*(\d+)\s*([%\w]+)?\s*$", value)
    if m is None:
        raise ValueError(f"{name}: invalid number/unit {value}")
    value = int(m.group(1))
    if m.group(2) is None:
        return value
    unit = m.group(2)[0]
    if unit not in units:
        valid = ", ".join(units.keys())
        raise ValueError(f"{name}: invalid unit '{unit}', valid values are {valid}")
    return value * units[unit]


def as_number(
    value: str | int, name: str | None = None, units: dict[str, int] | None = None, none_ok: bool = False
) -> int | None:
    """Deprecated function to convert a human readable string to a number.

    Parameters
    ----------
    value : str or int
        The value to convert
    name : str, optional
        The name of the value
    units : dict, optional
        The units to use for conversion
    none_ok : bool, optional
        Whether None is an acceptable value

    Returns
    -------
    int or None
        The converted value
    """
    warnings.warn(
        "Function as_number is deprecated and will be removed in a future version. Use human_to_number instead.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    return human_to_number(value, name, units, none_ok)


def human_seconds(value: str | int, name: str | None = None, none_ok: bool = False) -> int | None:
    """Convert a human readable string to seconds.

    Parameters
    ----------
    value : str or int
        The value to convert
    name : str, optional
        The name of the value
    none_ok : bool, optional
        Whether None is an acceptable value

    Returns
    -------
    int or None
        The converted value in seconds
    """
    units = dict(s=1, m=60, h=3600, d=86400, w=86400 * 7)
    return human_to_number(value, name, units, none_ok)


def as_seconds(value: str | int, name: str | None = None, none_ok: bool = False) -> int | None:
    """Deprecated function to convert a human readable string to seconds.

    Parameters
    ----------
    value : str or int
        The value to convert
    name : str, optional
        The name of the value
    none_ok : bool, optional
        Whether None is an acceptable value

    Returns
    -------
    int or None
        The converted value in seconds
    """
    warnings.warn(
        "Function as_seconds is deprecated and will be removed in a future version. Use human_seconds instead.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    return human_seconds(value, name, none_ok)


def human_to_percent(value: str | int, name: str | None = None, none_ok: bool = False) -> int | None:
    """Convert a human readable string to a percentage.

    Parameters
    ----------
    value : str or int
        The value to convert
    name : str, optional
        The name of the value
    none_ok : bool, optional
        Whether None is an acceptable value

    Returns
    -------
    int or None
        The converted value in percentage
    """
    units = {"%": 1}
    return human_to_number(value, name, units, none_ok)


def as_percent(value: str | int, name: str | None = None, none_ok: bool = False) -> int | None:
    """Deprecated function to convert a human readable string to a percentage.

    Parameters
    ----------
    value : str or int
        The value to convert
    name : str, optional
        The name of the value
    none_ok : bool, optional
        Whether None is an acceptable value

    Returns
    -------
    int or None
        The converted value in percentage
    """
    warnings.warn(
        "Function as_percent is deprecated and will be removed in a future version. Use human_to_percent instead.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    return human_to_percent(value, name, none_ok)


def human_to_bytes(value: str | int, name: str | None = None, none_ok: bool = False) -> int | None:
    """Convert a human readable string to bytes.

    Parameters
    ----------
    value : str or int
        The value to convert
    name : str, optional
        The name of the value
    none_ok : bool, optional
        Whether None is an acceptable value

    Returns
    -------
    int or None
        The converted value in bytes
    """
    units = {}
    n = 1
    for u in "KMGTP":
        n *= 1024
        units[u] = n
        units[u.lower()] = n

    return human_to_number(value, name, units, none_ok)


def as_bytes(value: str | int, name: str | None = None, none_ok: bool = False) -> int | None:
    """Deprecated function to convert a human readable string to bytes.

    Parameters
    ----------
    value : str or int
        The value to convert
    name : str, optional
        The name of the value
    none_ok : bool, optional
        Whether None is an acceptable value

    Returns
    -------
    int or None
        The converted value in bytes
    """
    warnings.warn(
        "Function as_bytes is deprecated and will be removed in a future version. Use human_to_bytes instead.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    return human_to_bytes(value, name, none_ok)


def human_to_timedelta(value: str, name: str | None = None, none_ok: bool = False) -> datetime.timedelta:
    """Convert a human readable string to a timedelta.

    Parameters
    ----------
    value : str
        The value to convert
    name : str, optional
        The name of the value
    none_ok : bool, optional
        Whether None is an acceptable value

    Returns
    -------
    datetime.timedelta
        The converted value as a timedelta
    """
    if value is None and none_ok:
        return None

    save = value
    value = re.sub(r"[^a-zA-Z0-9]", "", value.lower())
    value = re.sub(r"([a-zA-Z])[a-zA-Z]*", r"\1", value)
    # value = re.sub(r"[^dmhsw0-9]", "", value)
    bits = [b for b in re.split(r"([dmhsw])", value) if b != ""]

    times = defaultdict(int)

    val = None

    for i, n in enumerate(bits):
        if i % 2 == 0:
            val = int(n)
        else:
            assert n in ("d", "m", "h", "s", "w")
            times[n] = val
            val = None

    if val is not None:
        if name:
            raise ValueError(f"{name}: invalid period '{save}'")
        raise ValueError(f"Invalid period '{save}'")

    return datetime.timedelta(
        weeks=times["w"],
        days=times["d"],
        hours=times["h"],
        minutes=times["m"],
        seconds=times["s"],
    )


def as_timedelta(value: str, name: str | None = None, none_ok: bool = False) -> datetime.timedelta:
    """Deprecated function to convert a human readable string to a timedelta.

    Parameters
    ----------
    value : str
        The value to convert
    name : str, optional
        The name of the value
    none_ok : bool, optional
        Whether None is an acceptable value

    Returns
    -------
    datetime.timedelta
        The converted value as a timedelta
    """
    warnings.warn(
        "Function as_timedelta is deprecated and will be removed in a future version. Use human_to_timedelta instead.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    return human_to_timedelta(value, name, none_ok)


def rounded_datetime(d: datetime.datetime) -> datetime.datetime:
    """Round a datetime to the nearest second.

    Parameters
    ----------
    d : datetime.datetime
        The datetime to round

    Returns
    -------
    datetime.datetime
        The rounded datetime
    """
    if float(d.microsecond) / 1000.0 / 1000.0 >= 0.5:
        d = d + datetime.timedelta(seconds=1)
    d = d.replace(microsecond=0)
    return d


def json_pretty_dump(obj: Any, max_line_length: int = 120, default: Callable = str) -> str:
    """Custom JSON dump function that keeps dicts and lists on one line if they are short enough.

    Parameters
    ----------
    obj : Any
        The object to be dumped as JSON.
    max_line_length : int, optional
        Maximum allowed line length for pretty-printing. Default is 120.
    default : function, optional
        Default function to convert non-serializable objects. Default is str.

    Returns
    -------
    str
        JSON string.
    """

    def _format_json(obj: Any, indent_level: int = 0) -> str:
        """Helper function to format JSON objects with custom pretty-print rules.

        Parameters
        ----------
        obj : Any
            The object to format.
        indent_level : int, optional
            Current indentation level. Default is 0.

        Returns
        -------
        str
            Formatted JSON string.
        """
        indent = " " * 4 * indent_level
        if isinstance(obj, dict):
            items = []
            for key, value in obj.items():
                items.append(f'"{key}": {_format_json(value, indent_level + 1)}')
            line = "{" + ", ".join(items) + "}"
            if len(line) <= max_line_length:
                return line
            else:
                return "{\n" + ",\n".join([f"{indent}    {item}" for item in items]) + "\n" + indent + "}"
        elif isinstance(obj, list):
            items = [_format_json(item, indent_level + 1) for item in obj]
            line = "[" + ", ".join(items) + "]"
            if len(line) <= max_line_length:
                return line
            else:
                return "[\n" + ",\n".join([f"{indent}    {item}" for item in items]) + "\n" + indent + "]"
        else:
            return json.dumps(obj, default=default)

    return _format_json(obj)


def shorten_list(lst: list[Any] | tuple[Any], max_length: int = 5) -> list[Any] | tuple[Any]:
    """Shorten a list to a maximum length.

    Parameters
    ----------
    lst : list or tuple
        The list to be shortened.
    max_length : int, optional
        Maximum length of the shortened list. Default is 5.

    Returns
    -------
    list or tuple
        Shortened list.
    """
    if len(lst) <= max_length:
        return lst
    else:
        half = max_length // 2
        result = list(lst[:half]) + ["..."] + list(lst[max_length - half :])
        if isinstance(lst, tuple):
            return tuple(result)
        return result


def _compress_dates(
    dates: list[datetime.datetime],
) -> Generator[list[datetime.datetime] | tuple[datetime.datetime, datetime.datetime, datetime.timedelta], None, None]:
    """Compress a list of dates into a more compact representation.

    Parameters
    ----------
    dates : list of datetime.datetime
        The list of dates to compress

    Returns
    -------
    list or tuple
        The compressed dates
    """
    dates = sorted(dates)
    if len(dates) < 3:
        yield dates
        return

    prev = first = dates.pop(0)
    curr = dates.pop(0)
    delta = curr - prev
    while curr - prev == delta:
        prev = curr
        if not dates:
            break
        curr = dates.pop(0)

    yield (first, prev, delta)
    if dates:
        yield from _compress_dates([curr] + dates)


def compress_dates(dates: list[datetime.datetime | str]) -> str:
    """Compress a list of dates into a human-readable format.

    Parameters
    ----------
    dates : list
        A list of dates, as datetime objects or strings.

    Returns
    -------
    str
        A human-readable string representing the compressed dates.
    """

    dates = [as_datetime(_) for _ in dates]
    result = []

    for n in _compress_dates(dates):
        if isinstance(n, list):
            result.extend([str(_) for _ in n])
        else:
            result.append(" ".join([str(n[0]), "to", str(n[1]), "by", str(n[2])]))

    return result


def print_dates(dates: list[datetime.datetime | str]) -> None:
    """Print a list of dates in a human-readable format.

    Parameters
    ----------
    dates : list
        A list of dates, as datetime objects or strings.
    """
    print(compress_dates(dates))


def make_list_int(value: str | list[int] | tuple[int] | int) -> list[int]:
    """Convert a string like "1/2/3" or "1/to/3" or "1/to/10/by/2" to a list of integers.

    Parameters
    ----------
    value : str, list, tuple, int
        The value to convert to a list of integers.

    Returns
    -------
    list
        A list of integers.
    """
    if isinstance(value, str):
        if "/" not in value:
            return [int(value)]
        bits = value.split("/")
        if len(bits) == 3 and bits[1].lower() == "to":
            value = list(range(int(bits[0]), int(bits[2]) + 1, 1))

        elif len(bits) == 5 and bits[1].lower() == "to" and bits[3].lower() == "by":
            value = list(range(int(bits[0]), int(bits[2]) + int(bits[4]), int(bits[4])))

    if isinstance(value, list):
        return value
    if isinstance(value, tuple):
        return value
    if isinstance(value, int):
        return [value]

    raise ValueError(f"Cannot make list from {value}")



# (C) Copyright 2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from collections.abc import Iterator
from typing import Any

from pydantic import BaseModel as PydanticBaseModel
from pydantic import ValidationError
from pydantic_core import ErrorDetails

CUSTOM_MESSAGES = {
    "missing": "A config entry seems to be missing. If not please check for any typos.",
    "extra_forbidden": "Extra entries in the config are forebidden. Please check for typos.",
}


def convert_errors(e: ValidationError, custom_messages: dict[str, str]) -> list[ErrorDetails]:
    new_errors: list[ErrorDetails] = []
    for error in e.errors():
        custom_message = custom_messages.get(error["type"])

        if custom_message:

            ctx = error.get("ctx")
            error["msg"] = custom_message.format(**ctx) if ctx else custom_message
        new_errors.append(error)
    return new_errors


class ValidationError(Exception):
    pass


def allowed_values(v: Any, values: list[Any]) -> Any:
    if v not in values:
        msg = {f"Value {v} not in {values}"}
        raise ValidationError(msg)
    return v


def required_fields(model: type[PydanticBaseModel], recursive: bool = False) -> Iterator[str]:
    for name, field in model.model_fields.items():
        if not field.is_required():
            continue
        t = field.annotation
        if recursive and isinstance(t, type) and issubclass(t, PydanticBaseModel):
            yield from required_fields(t, recursive=True)
        else:
            yield name



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import json
import sys
from argparse import ArgumentParser
from argparse import Namespace

from anemoi.utils.mars.requests import print_request

from . import Command


class Requests(Command):
    """Convert a JSON requests file to MARS format."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.add_argument("input")
        command_parser.add_argument("output")
        command_parser.add_argument("--verb", default="retrieve")
        command_parser.add_argument("--only-one-field", action="store_true")

    def run(self, args: Namespace) -> None:
        """Execute the command with the provided arguments.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        if args.input == "-":
            requests = json.load(sys.stdin)
        else:
            with open(args.input) as f:
                requests = json.load(f)

        if args.only_one_field:
            for r in requests:
                for key in (
                    "grid",
                    "area",
                ):
                    r.pop(key, None)
                for k, v in list(r.items()):
                    if isinstance(v, list):
                        r[k] = v[-1]

        with open(args.output, "w") as f:
            for r in requests:
                print_request(args.verb, r, file=f)


command = Requests



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import concurrent.futures
import logging
import os
import shutil
from abc import abstractmethod
from collections.abc import Iterable
from typing import Any

import tqdm

from ..humanize import bytes_to_human

LOGGER = logging.getLogger(__name__)


def robust(call: callable, *args, maximum_tries: int = 60, retry_after: int = 60, **kwargs) -> callable:
    """Forwards the arguments to the multiurl robust function.
    with default retry_after=60 and maximum_tries=60.
    """
    from multiurl import robust as robust_

    return robust_(call, *args, retry_after=retry_after, maximum_tries=maximum_tries, **kwargs)


def _ignore(number_of_files: int, total_size: int, total_transferred: int, transfering: bool) -> None:
    """A placeholder function for progress reporting.

    Parameters
    ----------
    number_of_files : int
        The number of files being transferred.
    total_size : int
        The total size of the files being transferred.
    total_transferred : int
        The total size of the files transferred so far.
    transfering : bool
        Whether the transfer is in progress.
    """
    pass


class Loader:
    def transfer_folder(
        self,
        *,
        source: str,
        target: str,
        overwrite: bool = False,
        resume: bool = False,
        verbosity: int = 1,
        threads: int = 1,
        progress: callable = None,
    ) -> None:
        """Transfer a folder from the source to the target location.

        Parameters
        ----------
        source : str
            The source folder path.
        target : str
            The target folder path.
        overwrite : bool, optional
            Whether to overwrite the target if it exists, by default False.
        resume : bool, optional
            Whether to resume the transfer if possible, by default False.
        verbosity : int, optional
            The verbosity level, by default 1.
        threads : int, optional
            The number of threads to use, by default 1.
        progress : callable, optional
            A callable for progress reporting, by default None.
        """
        assert verbosity == 1, verbosity

        if progress is None:
            progress = _ignore

        # from boto3.s3.transfer import TransferConfig
        # config = TransferConfig(use_threads=False)
        config = None
        with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
            try:
                if verbosity > 0:
                    LOGGER.info(f"{self.action} {source} to {target}")

                total_size = 0
                total_transferred = 0

                futures = []
                for name in self.list_source(source):

                    futures.append(
                        executor.submit(
                            self.transfer_file,
                            source=self.source_path(name, source),
                            target=self.target_path(name, source, target),
                            overwrite=overwrite,
                            resume=resume,
                            verbosity=verbosity - 1,
                            config=config,
                        )
                    )
                    total_size += self.source_size(name)

                    if len(futures) % 10000 == 0:

                        progress(len(futures), total_size, 0, False)

                        if verbosity > 0:
                            LOGGER.info(f"Preparing transfer, {len(futures):,} files... ({bytes_to_human(total_size)})")
                        done, _ = concurrent.futures.wait(
                            futures,
                            timeout=0.001,
                            return_when=concurrent.futures.FIRST_EXCEPTION,
                        )
                        # Trigger exceptions if any
                        for future in done:
                            future.result()

                number_of_files = len(futures)
                progress(number_of_files, total_size, 0, True)

                if verbosity > 0:
                    LOGGER.info(f"{self.action} {number_of_files:,} files ({bytes_to_human(total_size)})")
                    with tqdm.tqdm(total=total_size, unit="B", unit_scale=True, unit_divisor=1024) as pbar:
                        for future in concurrent.futures.as_completed(futures):
                            size = future.result()
                            pbar.update(size)
                            total_transferred += size
                            progress(number_of_files, total_size, total_transferred, True)
                else:
                    for future in concurrent.futures.as_completed(futures):
                        size = future.result()
                        total_transferred += size
                        progress(number_of_files, total_size, total_transferred, True)

            except Exception:
                executor.shutdown(wait=False, cancel_futures=True)
                raise

    def transfer_file(
        self,
        source: str,
        target: str,
        overwrite: bool,
        resume: bool,
        verbosity: int,
        threads: int = 1,
        progress: callable = None,
        config: dict = None,
    ) -> int:
        """Transfer a file from the source to the target location.

        Parameters
        ----------
        source : str
            The source file path.
        target : str
            The target file path.
        overwrite : bool
            Whether to overwrite the target if it exists.
        resume : bool
            Whether to resume the transfer if possible.
        verbosity : int
            The verbosity level.
        threads : int, optional
            The number of threads to use, by default 1.
        progress : callable, optional
            A callable for progress reporting, by default None.
        config : dict, optional
            Additional configuration options, by default None.

        Returns
        -------
        int
            The size of the transferred file.

        Raises
        ------
        Exception
            If an error occurs during the transfer.
        """
        try:
            return self._transfer_file(source, target, overwrite, resume, verbosity, threads=threads, config=config)
        except Exception as e:
            LOGGER.exception(f"Error transferring {source} to {target}")
            LOGGER.error(e)
            raise

    @abstractmethod
    def list_source(self, source: str) -> Iterable:
        """List the files in the source location.

        Parameters
        ----------
        source : str
            The source location.

        Returns
        -------
        Iterable
            An iterable of files in the source location.
        """
        raise NotImplementedError

    @abstractmethod
    def source_path(self, local_path: str, source: str) -> str:
        """Get the source path for a local file.

        Parameters
        ----------
        local_path : str
            The local file path.
        source : str
            The source location.

        Returns
        -------
        str
            The source path for the local file.
        """
        raise NotImplementedError

    @abstractmethod
    def target_path(self, source_path: str, source: str, target: str) -> str:
        """Get the target path for a source file.

        Parameters
        ----------
        source_path : str
            The source file path.
        source : str
            The source location.
        target : str
            The target location.

        Returns
        -------
        str
            The target path for the source file.
        """
        raise NotImplementedError

    @abstractmethod
    def source_size(self, local_path: str) -> int:
        """Get the size of a local file.

        Parameters
        ----------
        local_path : str
            The local file path.

        Returns
        -------
        int
            The size of the local file.
        """
        raise NotImplementedError

    @abstractmethod
    def copy(self, source: str, target: str, **kwargs) -> None:
        """Copy a file or folder from the source to the target location.

        Parameters
        ----------
        source : str
            The source location.
        target : str
            The target location.
        kwargs : dict
            Additional arguments for the transfer.
        """
        raise NotImplementedError

    @abstractmethod
    def get_temporary_target(self, target: str, pattern: str) -> str:
        """Get a temporary target path based on the given pattern.

        Parameters
        ----------
        target : str
            The original target path.
        pattern : str
            The pattern to format the temporary path.

        Returns
        -------
        str
            The temporary target path.
        """
        raise NotImplementedError

    @abstractmethod
    def rename_target(self, target: str, temporary_target: str) -> None:
        """Rename the target to a new target path.

        Parameters
        ----------
        target : str
            The original target path.
        temporary_target : str
            The new target path.
        """
        raise NotImplementedError


class BaseDownload(Loader):
    action = "Downloading"

    @abstractmethod
    def copy(self, source: str, target: str, **kwargs) -> None:
        """Copy a file or folder from the source to the target location.

        Parameters
        ----------
        source : str
            The source location.
        target : str
            The target location.
        kwargs : dict
            Additional arguments for the transfer.
        """
        raise NotImplementedError

    def get_temporary_target(self, target: str, pattern: str) -> str:
        """Get a temporary target path based on the given pattern.

        Parameters
        ----------
        target : str
            The original target path.
        pattern : str
            The pattern to format the temporary path.

        Returns
        -------
        str
            The temporary target path.
        """
        if pattern is None:
            return target
        dirname, basename = os.path.split(target)
        return pattern.format(dirname=dirname, basename=basename)

    def rename_target(self, target: str, new_target: str) -> None:
        """Rename the target to a new target path.

        Parameters
        ----------
        target : str
            The original target path.
        new_target : str
            The new target path.
        """
        os.rename(target, new_target)

    def delete_target(self, target: str) -> None:
        """Delete the target if it exists.

        Parameters
        ----------
        target : str
            The target path.
        """
        if os.path.exists(target):
            if os.path.isfile(target):
                LOGGER.info(f"Deleting file {target}")
                os.remove(target)
            else:
                LOGGER.info(f"Deleting directory {target}")
                shutil.rmtree(target)


class BaseUpload(Loader):
    action = "Uploading"

    def copy(self, source: str, target: str, **kwargs) -> None:
        """Copy a file or folder from the source to the target location.

        Parameters
        ----------
        source : str
            The source location.
        target : str
            The target location.
        kwargs : dict
            Additional arguments for the transfer.
        """
        if os.path.isdir(source):
            self.transfer_folder(source=source, target=target, **kwargs)
        else:
            self.transfer_file(source=source, target=target, **kwargs)

    def list_source(self, source: str) -> Iterable:
        """List the files in the source location.

        Parameters
        ----------
        source : str
            The source location.

        Returns
        -------
        Iterable
            An iterable of files in the source location.
        """
        for root, _, files in os.walk(source):
            for file in files:
                yield os.path.join(root, file)

    def source_path(self, local_path: str, source: str) -> str:
        """Get the source path for a local file.

        Parameters
        ----------
        local_path : str
            The local file path.
        source : str
            The source location.

        Returns
        -------
        str
            The source path for the local file.
        """
        return local_path

    def target_path(self, source_path: str, source: str, target: str) -> str:
        """Get the target path for a source file.

        Parameters
        ----------
        source_path : str
            The source file path.
        source : str
            The source location.
        target : str
            The target location.

        Returns
        -------
        str
            The target path for the source file.
        """
        relative_path = os.path.relpath(source_path, source)
        path = os.path.join(target, relative_path)
        return path

    def source_size(self, local_path: str) -> int:
        """Get the size of a local file.

        Parameters
        ----------
        local_path : str
            The local file path.

        Returns
        -------
        int
            The size of the local file.
        """
        return os.path.getsize(local_path)


class TransferMethodNotImplementedError(NotImplementedError):
    pass


class Transfer:
    """This is the internal API and should not be used directly. Use the transfer function instead."""

    TransferMethodNotImplementedError = TransferMethodNotImplementedError

    def __init__(
        self,
        *,
        source: str,
        target: str,
        overwrite: bool = False,
        resume: bool = False,
        verbosity: int = 1,
        threads: int = 1,
        progress: callable = None,
        temporary_target: bool = False,
    ):
        if target == ".":
            target = os.path.basename(source)
            if not target:
                target = os.path.basename(os.path.dirname(source))

        temporary_target: dict[Any, Any] = {
            False: None,
            True: "{dirname}-downloading/{basename}",
            "-tmp/*": "{dirname}-tmp/{basename}",
            "*-tmp": "{dirname}/{basename}-tmp",
            "tmp-*": "{dirname}/tmp-{basename}",
        }.get(temporary_target, temporary_target)
        assert temporary_target is None or isinstance(temporary_target, str), (type(temporary_target), temporary_target)

        self.source = source
        self.target = target
        self.overwrite = overwrite
        self.resume = resume
        self.verbosity = verbosity
        self.threads = threads
        self.progress = progress
        self.temporary_target = temporary_target

        cls = _find_transfer_class(self.source, self.target)
        self.loader = cls()

    def run(self) -> "Transfer":
        """Execute the transfer process.

        Returns
        -------
        Transfer
            The Transfer instance.
        """
        target = self.loader.get_temporary_target(self.target, self.temporary_target)
        if target != self.target:
            LOGGER.info(f"Using temporary target {target} to copy to {self.target}")

        if self.overwrite:
            # delete the target if it exists
            LOGGER.info(f"Deleting {self.target}")
            self.delete_target(target)

            # carefully delete the temporary target if it exists
            head, tail = os.path.split(self.target)
            head_, tail_ = os.path.split(target)
            if not head_.startswith(head) or tail not in tail_:
                LOGGER.info(f"{target} is too different from {self.target} to delete it automatically.")
            else:
                self.delete_target(target)

        self.loader.copy(
            self.source,
            target,
            overwrite=self.overwrite,
            resume=self.resume,
            verbosity=self.verbosity,
            threads=self.threads,
            progress=self.progress,
        )

        self.rename_target(target, self.target)

        return self

    def rename_target(self, target: str, new_target: str) -> None:
        """Rename the target to a new target path.

        Parameters
        ----------
        target : str
            The original target path.
        new_target : str
            The new target path.
        """
        if target != new_target:
            LOGGER.info(f"Renaming temporary target {target} into {self.target}")
            return self.loader.rename_target(target, new_target)

    def delete_target(self, target: str) -> None:
        """Delete the target if it exists.

        Parameters
        ----------
        target : str
            The target path.
        """
        return self.loader.delete_target(target)


def _find_transfer_class(source: str, target: str) -> type:
    """Find the appropriate transfer class based on the source and target locations.

    Parameters
    ----------
    source : str
        The source location.
    target : str
        The target location.

    Returns
    -------
    type
        The transfer class.

    Raises
    ------
    TransferMethodNotImplementedError
        If the transfer method is not implemented.
    """
    from_ssh = source.startswith("ssh://")
    into_ssh = target.startswith("ssh://")

    from_s3 = source.startswith("s3://")
    into_s3 = target.startswith("s3://")

    from_local = not from_ssh and not from_s3
    into_local = not into_ssh and not into_s3

    # check that exactly one source type and one target type is specified
    assert sum([into_ssh, into_local, into_s3]) == 1, (into_ssh, into_local, into_s3)
    assert sum([from_ssh, from_local, from_s3]) == 1, (from_ssh, from_local, from_s3)

    if from_local and into_ssh:  # local -> ssh
        from .ssh import RsyncUpload

        return RsyncUpload

    if from_s3 and into_local:  # local <- S3
        from .s3 import S3Download

        return S3Download

    if from_local and into_s3:  # local -> S3
        from .s3 import S3Upload

        return S3Upload

    raise TransferMethodNotImplementedError(f"Transfer from {source} to {target} is not implemented")


# This function is the main entry point for the transfer mechanism for the other anemoi packages
def transfer(
    source, target, *, overwrite=False, resume=False, verbosity=1, progress=None, threads=1, temporary_target=False
) -> Loader:
    """Transfer files or folders from the source to the target location.

    Parameters
    ----------
    source : str
        A path to a local file or folder or a URL to a file or a folder on S3.
        The url should start with 's3://'.
    target : str
        A path to a local file or folder or a URL to a file or a folder on S3 or a remote folder.
        The url should start with 's3://' or 'ssh://'.
    overwrite : bool, optional
        If the data is already on in the target location it will be overwritten.
        By default False
    resume : bool, optional
        If the data is already on S3 it will not be uploaded, unless the remote file has a different size
        Ignored if the target is an SSH remote folder (ssh://).
        By default False
    verbosity : int, optional
        The level of verbosity, by default 1
    progress : callable, optional
        A callable that will be called with the number of files, the total size of the files, the total size
        transferred and a boolean indicating if the transfer has started. By default None
    threads : int, optional
        The number of threads to use when uploading a directory, by default 1
    temporary_target : bool, optional
        Experimental feature
        If True and if the target location supports it, the data will be uploaded to a temporary location
        then renamed to the final location. Supported by SSH and local targets, not supported by S3.
        By default False.

    Returns
    -------
    Loader
        The Loader instance.
    """
    copier = Transfer(
        source=source,
        target=target,
        overwrite=overwrite,
        resume=resume,
        verbosity=verbosity,
        progress=progress,
        threads=threads,
        temporary_target=temporary_target,
    )
    copier.run()
    return copier



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from anemoi.utils.config import DotDict
from anemoi.utils.config import _merge_dicts
from anemoi.utils.config import _set_defaults
from anemoi.utils.grib import paramid_to_shortname
from anemoi.utils.grib import shortname_to_paramid


def test_dotdict() -> None:
    """Test the DotDict class for nested dictionary access and assignment.

    Tests:
        - Accessing nested dictionary values.
        - Assigning new values to existing keys.
        - Adding new nested dictionaries.
        - Accessing and assigning values in nested lists.
    """
    d = DotDict(a=1, b=2, c=dict(d=3, e=4), e=[1, dict(a=3), 3])
    assert d.a == 1
    assert d.b == 2
    assert d.c.d == 3
    assert d.c.e == 4

    d.a = 10
    assert d.a == 10

    d.d = dict(f=5)
    assert d.d.f == 5

    d.d.x = 6
    assert d.d.x == 6

    assert d.e[1].a == 3


def test_add_nested_dict_via_setitem():
    """Test that assigning a nested dict via item access results in recursive DotDict conversion."""

    d = DotDict(a=1)
    d["b"] = {
        "c": "d",
    }
    assert d.b.c == "d"


def test_adding_list_of_dicts_via_setitem():
    d = DotDict(a=1)
    d["b"] = [
        {
            "c": "d",
        },
        {"e": "f"},
    ]
    assert d.b[0].c == "d"
    assert d.b[1].e == "f"


def test_adding_list_of_dicts_via_setattr():
    d = DotDict(a=1)
    d.b = [
        {
            "c": "d",
        },
        {"e": "f"},
    ]
    assert d.b[0].c == "d"
    assert d.b[1].e == "f"


def test_merge_dicts() -> None:
    """Test the _merge_dicts function for merging nested dictionaries.

    Tests:
        - Merging two dictionaries with overlapping keys.
        - Ensuring nested dictionaries are merged correctly.
    """
    a = dict(a=1, b=2, c=dict(d=3, e=4))
    b = dict(a=10, c=dict(a=30, e=40), d=9)
    _merge_dicts(a, b)
    assert a == {"a": 10, "b": 2, "c": {"d": 3, "e": 40, "a": 30}, "d": 9}


def test_set_defaults() -> None:
    """Test the _set_defaults function for setting default values in nested dictionaries.

    Tests:
        - Setting default values without overwriting existing ones.
        - Ensuring nested dictionaries are handled correctly.
    """
    a = dict(a=1, b=2, c=dict(d=3, e=4))
    b = dict(a=10, c=dict(a=30, e=40), d=9)
    _set_defaults(a, b)
    assert a == {"a": 1, "b": 2, "c": {"d": 3, "e": 4, "a": 30}, "d": 9}


def test_grib() -> None:
    """Test the GRIB utility functions.

    Tests:
        - Converting short names to parameter IDs.
        - Converting parameter IDs to short names.
    """
    assert shortname_to_paramid("2t") == 167
    assert paramid_to_shortname(167) == "2t"


if __name__ == "__main__":
    """Run all test functions."""
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
from __future__ import annotations

import os

import requests

from ..remote import robust


def health_check(tracking_uri: str) -> None:
    """Query the health endpoint of an MLflow server.

    If the server is not reachable, raise an error and remind the user that authentication may be required.

    Raises
    ------
    ConnectionError
        If the server is not reachable.

    """
    token = os.getenv("MLFLOW_TRACKING_TOKEN")

    headers = {"Authorization": f"Bearer {token}"}
    response = robust(requests.get, retry_after=30, maximum_tries=10)(
        f"{tracking_uri}/health",
        headers=headers,
        timeout=60,
    )

    if response.text == "OK":
        return

    error_msg = f"Could not connect to MLflow server at {tracking_uri}. "
    if not token:
        error_msg += "The server may require authentication, did you forget to turn it on?"
    raise ConnectionError(error_msg)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from argparse import ArgumentParser
from argparse import Namespace

from anemoi.utils.remote import transfer

from . import Command


class Transfer(Command):
    """Transfer files or folders from the source to the target location."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.add_argument(
            "--source", help="A path to a local file or folder or a URL to a file or a folder on S3."
        )
        command_parser.add_argument(
            "--target", help="A path to a local file or folder or a URL to a file or a folder on S3 or a remote folder."
        )
        command_parser.add_argument(
            "--overwrite",
            action="store_true",
            help="If the data is already on in the target location it will be overwritten..",
        )
        command_parser.add_argument(
            "--resume",
            action="store_true",
            help="If the data is already on S3 it will not be uploaded, unless the remote file has a different size.",
        )
        command_parser.add_argument("--verbosity", default=1, help="The level of verbosity, by default 1.")
        command_parser.add_argument(
            "--progress", default=None, help="A callable that will be called with the number of files."
        )
        command_parser.add_argument(
            "--threads", default=1, help="The number of threads to use when uploading a directory, by default 1."
        )

    def run(self, args: Namespace) -> None:
        """Execute the command with the provided arguments.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        transfer(
            source=args.source,
            target=args.target,
            overwrite=args.overwrite,
            resume=args.resume,
            verbosity=args.verbosity,
            progress=args.progress,
            threads=args.threads,
            temporary_target=False,
        )


command = Transfer



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import numpy as np

from anemoi.utils.caching import cached
from anemoi.utils.caching import clean_cache


class Data(dict):
    """Simple class to store data and count the number of calls to the function.

    Attributes
    ----------
    n : int
        The number of calls to the function.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the Data object."""
        super().__init__(*args, **kwargs)
        self.n = 0


def check(f: callable, data: Data) -> None:
    """Check that the function f returns the expected values from the data.

    Parameters
    ----------
    f : callable
        The function to be tested.
    data : Data
        The data to be used for testing.
    """
    """
    The function f is called three times for each value in the data.
    The number of actual calls to the function is checked to make sure the cache is used when it should be.
    """

    for i, x in enumerate(data):
        assert data.n == i

        res = f(x)
        assert type(res) == type(data[x])  # noqa: E721
        assert str(res) == str(data[x])
        assert data.n == i + 1

        res = f(x)
        assert type(res) == type(data[x])  # noqa: E721
        assert str(res) == str(data[x])
        assert data.n == i + 1


global DATA

#########################################
# basic test
#########################################
global values_a
values_a = Data(a=1, b=2)


@cached(collection="test", expires=0)
def func_a(x: str) -> int:
    """Example function to be cached.

    Parameters
    ----------
    x : str
        The key to access the data.

    Returns
    -------
    int
        The value associated with the key.
    """
    global values_a
    values_a.n += 1
    return values_a[x]


def test_cached_basic(*values: str, **kwargs: dict) -> None:
    """Test the cached decorator with basic data types.

    Parameters
    ----------
    values : str
        The values to be tested.
    kwargs : dict
        Additional keyword arguments.
    """
    clean_cache("test")
    check(func_a, values_a)


#########################################
# Test with numpy arrays
#########################################

global values_c
values_c = Data(
    a=dict(A=np.array([1, 2, 3]), B=np.array([4, 5, 6])),
    b=dict(A=np.array([7, 8, 9]), B=np.array([10, 11, 12])),
)


@cached(collection="test", expires=0, encoding="npz")
def func_c(x: str) -> dict:
    """Example function to be cached with numpy arrays.

    Parameters
    ----------
    x : str
        The key to access the data.

    Returns
    -------
    dict
        The value associated with the key.
    """
    global values_c
    values_c.n += 1
    return values_c[x]


def test_cached_npz(*values: str, **kwargs: dict) -> None:
    """Test the cached decorator with numpy arrays.

    Parameters
    ----------
    values : str
        The values to be tested.
    kwargs : dict
        Additional keyword arguments.
    """
    clean_cache("test")
    check(func_c, values_c)


#########################################
# Test with a various types
global values_d
values_d = Data(a="4", b=5.0, c=dict(d=6), e=[7, 8, 9], f=(10, 11, 12))


@cached(collection="test", expires=0)
def func_d(x: str) -> any:
    """Example function to be cached with various data types.

    Parameters
    ----------
    x : str
        The key to access the data.

    Returns
    -------
    any
        The value associated with the key.
    """
    global values_d
    values_d.n += 1
    return values_d[x]


def test_cached_various_types(*values: str, **kwargs: dict) -> None:
    """Test the cached decorator with various data types.

    Parameters
    ----------
    values : str
        The values to be tested.
    kwargs : dict
        Additional keyword arguments.
    """
    clean_cache("test")
    check(func_d, values_d)


#########################################

if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from __future__ import annotations

import logging
import os
import time
from datetime import datetime
from datetime import timezone
from functools import wraps
from getpass import getpass
from typing import TYPE_CHECKING

import requests
from requests.exceptions import HTTPError

from ..config import config_path
from ..config import load_config
from ..config import save_config
from ..remote import robust
from ..timer import Timer

REFRESH_EXPIRE_DAYS = 29


if TYPE_CHECKING:
    from collections.abc import Callable


class TokenAuth:
    """Manage authentication with a keycloak token server."""

    config_file = "mlflow-token.json"

    def __init__(
        self,
        url: str,
        enabled: bool = True,
        target_env_var: str = "MLFLOW_TRACKING_TOKEN",
    ) -> None:
        """Initialise the token authentication object.

        Parameters
        ----------
        url : str
            URL of the authentication server.
        enabled : bool, optional
            Set this to False to turn off authentication, by default True
        target_env_var : str, optional
            The environment variable to store the access token in after authenticating,
            by default `MLFLOW_TRACKING_TOKEN`

        """
        self.url = url
        self.target_env_var = target_env_var
        self._enabled = enabled

        config = self.load_config()

        self._refresh_token = config.get("refresh_token")
        self.refresh_expires = config.get("refresh_expires", 0)
        self.access_token = None
        self.access_expires = 0

        # the command line tool adds a default handler to the root logger on runtime,
        # so we init our logger here (on runtime, not on import) to avoid duplicate handlers
        self.log = logging.getLogger(__name__)

    def __call__(self) -> None:
        self.authenticate()

    @property
    def refresh_token(self) -> str:
        return self._refresh_token

    @refresh_token.setter
    def refresh_token(self, value: str) -> None:
        self._refresh_token = value
        self.refresh_expires = time.time() + (REFRESH_EXPIRE_DAYS * 86400)  # 86400 seconds in a day

    @staticmethod
    def load_config() -> dict:
        path = config_path(TokenAuth.config_file)

        if not os.path.exists(path):
            save_config(TokenAuth.config_file, {})

        if os.path.exists(path) and os.stat(path).st_mode & 0o777 != 0o600:
            os.chmod(path, 0o600)

        return load_config(TokenAuth.config_file)

    def enabled(fn: Callable) -> Callable:  # noqa: N805
        """Decorator to call or ignore a function based on the `enabled` flag."""

        @wraps(fn)
        def _wrapper(self: TokenAuth, *args, **kwargs) -> Callable | None:
            if self._enabled:
                return fn(self, *args, **kwargs)
            return None

        return _wrapper

    @enabled
    def login(self, force_credentials: bool = False, **kwargs: dict) -> None:
        """Acquire a new refresh token and save it to disk.

        If an existing valid refresh token is already on disk it will be used.
        If not, or the token has expired, the user will be asked to obtain one from the API.

        Refresh token expiry time is set in the `REFRESH_EXPIRE_DAYS` constant (default 29 days).

        This function should be called once, interactively, right before starting a training run.

        Parameters
        ----------
        force_credentials : bool, optional
            Force a credential login even if a refreh token is available, by default False.
        kwargs : dict
            Additional keyword arguments.

        Raises
        ------
        RuntimeError
            A new refresh token could not be acquired.

        """
        del kwargs  # unused
        self.log.info("🌐 Logging in to %s", self.url)
        new_refresh_token = None

        if not force_credentials and self.refresh_token and self.refresh_expires > time.time():
            new_refresh_token = self._token_request(ignore_exc=True).get("refresh_token")

        if not new_refresh_token:
            self.log.info("📝 Please obtain a seed refresh token from %s/seed", self.url)
            self.log.info("📝 and paste it here (you will not see the output, just press enter after pasting):")
            self.refresh_token = getpass("Refresh Token: ")

            # perform a new refresh token request to check if the seed refresh token is valid
            new_refresh_token = self._token_request().get("refresh_token")

        if not new_refresh_token:
            msg = "❌ Failed to log in. Please try again."
            raise RuntimeError(msg)

        self.refresh_token = new_refresh_token
        self.save()

        self.log.info("✅ Successfully logged in to MLflow. Happy logging!")

    @enabled
    def authenticate(self, **kwargs: dict) -> None:
        """Check the access token and refresh it if necessary. A new refresh token will also be acquired upon refresh.

        This requires a valid refresh token to be available, obtained from the `login` method.

        The access token is stored in memory and in an environment variable.
        If the access token is still valid, this function does nothing.

        This function should be called before every MLflow API request.

        Raises
        ------
        RuntimeError
            No refresh token is available or the token request failed.

        """
        del kwargs  # unused
        if self.access_expires > time.time():
            return

        if not self.refresh_token or self.refresh_expires < time.time():
            msg = "You are not logged in to MLflow. Please log in first."
            raise RuntimeError(msg)

        with Timer("Access token refreshed", self.log):
            response = self._token_request()

        self.access_token = response.get("access_token")
        self.access_expires = time.time() + (response.get("expires_in") * 0.7)  # bit of buffer
        self.refresh_token = response.get("refresh_token")

        os.environ[self.target_env_var] = self.access_token

    @enabled
    def save(self, **kwargs: dict) -> None:
        """Save the latest refresh token to disk."""
        del kwargs  # unused
        if not self.refresh_token:
            self.log.warning("No refresh token to save.")
            return

        config = {
            "url": self.url,
            "refresh_token": self.refresh_token,
            "refresh_expires": self.refresh_expires,
        }
        save_config(self.config_file, config)

        expire_date = datetime.fromtimestamp(self.refresh_expires, tz=timezone.utc)
        self.log.info(
            "Your MLflow login token is valid until %s UTC",
            expire_date.strftime("%Y-%m-%d %H:%M:%S"),
        )

    def _token_request(
        self,
        ignore_exc: bool = False,
    ) -> dict:
        path = "refreshtoken"
        payload = {"refresh_token": self.refresh_token}

        try:
            response = self._request(path, payload)
        except Exception:
            if ignore_exc:
                return {}
            raise

        return response

    def _request(self, path: str, payload: dict) -> dict:

        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }

        try:
            response = robust(requests.post)(
                f"{self.url}/{path}",
                headers=headers,
                json=payload,
                timeout=60,
            )
            response.raise_for_status()
            response_json = response.json()

            if response_json.get("status", "") != "OK":
                error_description = response_json.get("response", "Error acquiring token.")
                msg = f"❌ {error_description}"
                raise RuntimeError(msg)

            return response_json["response"]
        except HTTPError:
            self.log.exception("HTTP error occurred")
            raise



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
from argparse import ArgumentParser
from argparse import Namespace

from ..config import config_path
from ..config import load_config
from . import Command


class Config(Command):
    """Handle configuration related commands."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.add_argument("--path", help="Print path to config file")

    def run(self, args: Namespace) -> None:
        """Execute the command with the provided arguments.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        if args.path:
            print(config_path())
        else:
            print(json.dumps(load_config(), indent=4))


command = Config



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


"""Text utilities."""

import re
from collections import defaultdict
from typing import Any
from typing import Optional

# https://en.wikipedia.org/wiki/Box-drawing_character


def dotted_line(width=84) -> str:
    """Return a dotted line using '┈'.

    >>> dotted_line(40)
    ┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈

    Parameters
    ----------
    width : int, optional
        Number of characters, by default 84

    Returns
    -------
    str
        The dotted line
    """

    return "┈" * width


# Regular expression to match ANSI escape codes
_ansi_escape = re.compile(r"\x1b\[([0-9;]*[mGKH])")


def _has_ansi_escape(s: str) -> bool:
    """Check if a string contains ANSI escape codes.

    Parameters
    ----------
    s : str
        The string to check.

    Returns
    -------
    bool
        True if the string contains ANSI escape codes, False otherwise.
    """
    return _ansi_escape.search(s) is not None


def _split_tokens(s: str) -> list[tuple[str, int]]:
    """Split a string into a list of visual characters with their lengths.

    Parameters
    ----------
    s : str
        The string to split.

    Returns
    -------
    list of tuple
        A list of tuples where each tuple contains a visual character and its length.
    """
    from wcwidth import wcswidth

    initial = s
    out = []

    # Function to probe the number of bytes needed to encode the first character
    def probe_utf8(s):
        for i in range(1, 5):
            try:
                s[:i].encode("utf-8")
            except UnicodeEncodeError:
                return i - 1
        return 1

    while s:
        match = _ansi_escape.match(s)
        if match:
            token = match.group(0)
            s = s[len(token) :]
            out.append((token, 0))
        else:
            i = probe_utf8(s)
            token = s[:i]
            s = s[i:]
            out.append((token, wcswidth(token)))

    assert "".join(token for (token, _) in out) == initial, (out, initial)
    return out


def visual_len(s: str | list[tuple[str, int]]) -> int:
    """Compute the length of a string as it appears on the terminal.

    Parameters
    ----------
    s : str or list of tuple
        The string or list of visual characters with their lengths.

    Returns
    -------
    int
        The visual length of the string.
    """
    if isinstance(s, str):
        s = _split_tokens(s)
    assert isinstance(s, (tuple, list)), (type(s), s)
    if len(s) == 0:
        return 0
    for _ in s:
        assert isinstance(_, tuple), s
        assert len(_) == 2, s
    n = 0
    for _, width in s:
        n += width
    return n


def boxed(text: str, min_width: int = 80, max_width: int = 80) -> str:
    """Put a box around a text.

    >>> boxed("Hello,\\nWorld!", max_width=40)
    ┌──────────────────────────────────────────┐
    │ Hello,                                   │
    │ World!                                   │
    └──────────────────────────────────────────┘

    Parameters
    ----------
    text : str
        The text to box
    min_width : int, optional
        The minimum width of the box, by default 80
    max_width : int, optional
        The maximum width of the box, by default 80

    Returns
    -------
    str
        A boxed version of the input text
    """

    lines = []
    for line in text.split("\n"):
        line = line.rstrip()
        line = _split_tokens(line)
        lines.append(line)

    width = max(visual_len(_) for _ in lines)

    if min_width is not None:
        width = max(width, min_width)

    if max_width is not None:

        def shorten_line(line, max_width):
            if visual_len(line) > max_width:
                while visual_len(line) >= max_width:
                    line = line[:-1]
                line.append(("…", 1))
            return line

        width = min(width, max_width)
        lines = [shorten_line(line, max_width) for line in lines]

    def pad_line(line, width):
        line = line + [" "] * (width - visual_len(line))
        return line

    lines = [pad_line(line, width) for line in lines]

    box = []
    box.append("┌" + "─" * (width + 2) + "┐")
    for line in lines:
        s = "".join(_[0] for _ in line)
        if _has_ansi_escape(s):
            s += "\x1b[0m"
        box.append(f"│ {s} │")
    box.append("└" + "─" * (width + 2) + "┘")

    return "\n".join(box)


def bold(text: str) -> str:
    """Make the text bold.

    Parameters
    ----------
    text : str
        The text to make bold.

    Returns
    -------
    str
        The bold text.
    """
    from termcolor import colored

    return colored(text, attrs=["bold"])


def red(text: str) -> str:
    """Make the text red.

    Parameters
    ----------
    text : str
        The text to make red.

    Returns
    -------
    str
        The red text.
    """
    from termcolor import colored

    return colored(text, "red")


def green(text: str) -> str:
    """Make the text green.

    Parameters
    ----------
    text : str
        The text to make green.

    Returns
    -------
    str
        The green text.
    """
    from termcolor import colored

    return colored(text, "green")


def blue(text: str) -> str:
    """Make the text blue.

    Parameters
    ----------
    text : str
        The text to make blue.

    Returns
    -------
    str
        The blue text.
    """
    from termcolor import colored

    return colored(text, "blue")


class Tree:
    """Tree data structure.

    Parameters
    ----------
    actor : Any
        The actor associated with the tree node.
    parent : Tree, optional
        The parent tree node, by default None.
    """

    def __init__(self, actor: Any, parent: Optional["Tree"] = None):
        self._actor = actor
        self._kids = []
        self._parent = parent

    def adopt(self, kid: "Tree") -> None:
        """Adopt a child tree node.

        Parameters
        ----------
        kid : Tree
            The child tree node to adopt.
        """
        kid._parent._kids.remove(kid)
        self._kids.append(kid)
        kid._parent = self
        # assert False

    def forget(self) -> None:
        """Forget the current tree node."""
        self._parent._kids.remove(self)
        self._parent = None

    @property
    def is_leaf(self) -> bool:
        """Bool: True if the tree node is a leaf, False otherwise."""
        return len(self._kids) == 0

    @property
    def key(self) -> tuple:
        """Tuple: The key of the tree node."""
        return tuple(sorted(self._actor.as_dict().items()))

    @property
    def _text(self) -> str:
        """Str: The text representation of the tree node."""
        return self._actor.summary

    @property
    def summary(self) -> str:
        """Str: The summary of the tree node."""
        return self._actor.summary

    def as_dict(self) -> dict:
        """Convert the tree node to a dictionary.

        Returns
        -------
        dict
            The dictionary representation of the tree node.
        """
        return self._actor.as_dict()

    def node(self, actor: Any, insert: bool = False) -> "Tree":
        """Create a new tree node.

        Parameters
        ----------
        actor : Any
            The actor associated with the new tree node.
        insert : bool, optional
            Whether to insert the new tree node at the beginning, by default False.

        Returns
        -------
        Tree
            The new tree node.
        """
        node = Tree(actor, self)
        if insert:
            self._kids.insert(0, node)
        else:
            self._kids.append(node)
        return node

    def print(self) -> None:
        """Print the tree."""
        padding = []

        while self._factorise():
            pass

        self._print(padding)

    def _leaves(self, result: list["Tree"]) -> None:
        """Collect all leaf nodes.

        Parameters
        ----------
        result : list of Tree
            The list to collect the leaf nodes.
        """
        if self.is_leaf:
            result.append(self)
        else:
            for kid in self._kids:
                kid._leaves(result)

    def _factorise(self) -> bool:
        """Factorise the tree.

        Returns
        -------
        bool
            True if the tree was factorised, False otherwise.
        """
        if len(self._kids) == 0:
            return False

        result = False
        for kid in self._kids:
            result = kid._factorise() or result

        if result:
            return True

        same = defaultdict(list)
        for kid in self._kids:
            for grand_kid in kid._kids:
                same[grand_kid.key].append((kid, grand_kid))

        result = False
        n = len(self._kids)
        texts = []
        for text, v in same.items():
            if len(v) == n and n > 1:
                for kid, grand_kid in v:
                    kid._kids.remove(grand_kid)
                texts.append((text, v[1][1]))
                result = True

        for text, actor in reversed(texts):
            self.node(actor, True)

        if result:
            return True

        if len(self._kids) != 1:
            return False

        kid = self._kids[0]
        texts = []
        for grand_kid in list(kid._kids):
            if len(grand_kid._kids) == 0:
                kid._kids.remove(grand_kid)
                texts.append((grand_kid.key, grand_kid))
                result = True

        for text, actor in reversed(texts):
            self.node(actor, True)

        return result

    def _print(self, padding: list[str]) -> None:
        """Print the tree with padding.

        Parameters
        ----------
        padding : list of str
            The padding for each level of the tree.
        """
        for i, p in enumerate(padding[:-1]):
            if p == " └":
                padding[i] = "  "
            if p == " ├":
                padding[i] = " │"
        if padding:
            print(f"{''.join(padding)}─{self._text}")
        else:
            print(self._text)
        padding.append(" ")
        for i, k in enumerate(self._kids):
            sep = " ├" if i < len(self._kids) - 1 else " └"
            padding[-1] = sep
            k._print(padding)

        padding.pop()

    def to_json(self, depth: int = 0) -> dict:
        """Convert the tree to a JSON serializable dictionary.

        Parameters
        ----------
        depth : int, optional
            The depth of the tree, by default 0.

        Returns
        -------
        dict
            The JSON serializable dictionary representation of the tree.
        """
        while self._factorise():
            pass

        return {
            "actor": self._actor.as_dict(),
            "kids": [k.to_json(depth + 1) for k in self._kids],
            "depth": depth,
        }


def table(rows: list[list[Any]], header: list[str], align: list[str], margin: int = 0) -> str:
    """Format a table.

        >>> table([['Aa', 12, 5],
                ['B', 120, 1],
                ['C', 9, 123]],
                ['C1', 'C2', 'C3'],
                ['<', '>', '>'])
            C1 │  C2 │  C3
            ───┼─────┼────
            Aa │  12 │   5
            B  │ 120 │   1
            C  │   9 │ 123
            ───┴─────┴────

    Parameters
    ----------
    rows : list of lists (or tuples)
        The rows of the table
    header : A list or tuple of strings
        The header of the table
    align : A list of '<', '>', or '^'
        To align the columns to the left, right, or center
    margin : int, optional
        Extra spaces on the left side of the table, by default 0

    Returns
    -------
    str
        A table as a string
    """

    def _(x):
        try:
            x = float(x)
        except Exception:
            pass

        if isinstance(x, float):
            return f"{x:g}"

        if isinstance(x, str):
            return x
        if isinstance(x, int):
            return str(x)

        return str(x)

    tmp = []
    for row in rows:
        tmp.append([_(x) for x in row])

    all_rows = [header] + tmp

    lens = [max(len(x) for x in col) for col in zip(*all_rows)]

    result = []
    for i, row in enumerate(all_rows):

        def _(x, i, j):
            if align[j] == "<":
                return x.ljust(i)
            if align[j] == ">":
                return x.rjust(i)
            return x.center(i)

        result.append(" │ ".join([_(x, i, j) for j, (x, i) in enumerate(zip(row, lens))]))
        if i == 0:
            result.append("─┼─".join(["─" * i for i in lens]))

    result.append("─┴─".join(["─" * i for i in lens]))

    if margin:
        result = [margin * " " + x for x in result]

    return "\n".join(result)


def progress(done: int, todo: int, width: int = 80) -> str:
    """Generates a progress bar string.

    Parameters
    ----------

    done : int
        The number of tasks completed.
    todo : int
        The total number of tasks.
    width : int, optional
        The width of the progress bar, by default 80.

    Returns
    -------
    str
        A string representing the progress bar.

    Example
    -------

    >>> print(progress(10, 100,width=50))
    █████▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒
    """
    done = min(int(done / todo * width + 0.5), width)
    return green("█" * done) + red("█" * (width - done))



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from anemoi.utils.sanitise import sanitise


def test_sanitise_urls() -> None:
    """Test the sanitise function for sanitizing URLs with sensitive information."""
    assert sanitise("http://johndoe:password@host:port/path") == "http://user:***@host:port/path"

    assert sanitise("http://www.example.com/path?pass=secret") == "http://www.example.com/path?pass=hidden"
    assert sanitise("http://www.example.com/path?password=secret") == "http://www.example.com/path?password=hidden"
    assert sanitise("http://www.example.com/path?token=secret") == "http://www.example.com/path?token=hidden"
    assert sanitise("http://www.example.com/path?user=secret") == "http://www.example.com/path?user=hidden"
    assert sanitise("http://www.example.com/path?key=secret") == "http://www.example.com/path?key=hidden"
    assert sanitise("http://www.example.com/path?pwd=secret") == "http://www.example.com/path?pwd=hidden"
    assert sanitise("http://www.example.com/path?_key=secret") == "http://www.example.com/path?_key=hidden"
    assert sanitise("http://www.example.com/path?_token=secret") == "http://www.example.com/path?_token=hidden"
    assert sanitise("http://www.example.com/path?apikey=secret") == "http://www.example.com/path?apikey=hidden"
    assert sanitise("http://www.example.com/path?api_key=secret") == "http://www.example.com/path?api_key=hidden"
    assert sanitise("http://www.example.com/path?api_token=secret") == "http://www.example.com/path?api_token=hidden"
    assert sanitise("http://www.example.com/path?_api_token=secret") == "http://www.example.com/path?_api_token=hidden"
    assert sanitise("http://www.example.com/path?_api_key=secret") == "http://www.example.com/path?_api_key=hidden"
    assert sanitise("http://www.example.com/path?username=secret") == "http://www.example.com/path?username=hidden"
    assert sanitise("http://www.example.com/path?login=secret") == "http://www.example.com/path?login=hidden"

    assert sanitise("http://www.example.com/path;pass=secret") == "http://www.example.com/path;pass=hidden"
    assert sanitise("http://www.example.com/path;password=secret") == "http://www.example.com/path;password=hidden"
    assert sanitise("http://www.example.com/path;token=secret") == "http://www.example.com/path;token=hidden"
    assert sanitise("http://www.example.com/path;user=secret") == "http://www.example.com/path;user=hidden"
    assert sanitise("http://www.example.com/path;key=secret") == "http://www.example.com/path;key=hidden"
    assert sanitise("http://www.example.com/path;pwd=secret") == "http://www.example.com/path;pwd=hidden"
    assert sanitise("http://www.example.com/path;_key=secret") == "http://www.example.com/path;_key=hidden"
    assert sanitise("http://www.example.com/path;_token=secret") == "http://www.example.com/path;_token=hidden"
    assert sanitise("http://www.example.com/path;apikey=secret") == "http://www.example.com/path;apikey=hidden"
    assert sanitise("http://www.example.com/path;api_key=secret") == "http://www.example.com/path;api_key=hidden"
    assert sanitise("http://www.example.com/path;api_token=secret") == "http://www.example.com/path;api_token=hidden"
    assert sanitise("http://www.example.com/path;_api_token=secret") == "http://www.example.com/path;_api_token=hidden"
    assert sanitise("http://www.example.com/path;_api_key=secret") == "http://www.example.com/path;_api_key=hidden"
    assert sanitise("http://www.example.com/path;username=secret") == "http://www.example.com/path;username=hidden"
    assert sanitise("http://www.example.com/path;login=secret") == "http://www.example.com/path;login=hidden"

    assert sanitise("http://www.example.com/path;_api_token=secret", level=2) == "http://hidden/path;_api_token=hidden"
    assert sanitise("http://johndoe:password@host:port/path", level=2) == "http://user:***@hidden/path"
    assert sanitise("http://host:port/path", level=2) == "http://hidden/path"

    assert sanitise("http://www.example.com/path;_api_token=secret", level=3) == "http://hidden"


def test_sanitise_paths() -> None:
    """Test the sanitise function for sanitizing file paths."""
    # We want to keep earthkit-data's url and path pattern

    assert sanitise("/home/johndoe/.ssh/id_rsa") == "/.../id_rsa"

    assert (
        sanitise("/data/model/{date:strftime(%Y)}/{date:strftime(%m)}/{date:strftime(%d)}/analysis.grib")
        == "/.../{date:strftime(%Y)}/{date:strftime(%m)}/{date:strftime(%d)}/analysis.grib"
    )

    assert sanitise("test.grib") == "test.grib"
    assert sanitise("../test.grib") == "../test.grib"
    assert sanitise("./test.grib") == "./test.grib"
    assert sanitise("sub/folder/test.grib") == "sub/folder/test.grib"
    assert sanitise("./folder/test.grib") == "./folder/test.grib"

    assert sanitise("./folder/test.grib", level=2) == "./folder/test.grib"

    assert sanitise("./folder/test.grib", level=3) == "hidden"
    assert sanitise("/home/johndoe/.ssh/id_rsa", level=3) == "hidden"


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024-2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


"""This module provides functions to upload, download, list and delete files and folders on S3.
The functions of this package expect that the AWS credentials are set up in the environment
typicaly by setting the `AWS_ACCESS_KEY_ID` and `AWS_SECRET_ACCESS_KEY` environment variables or
by creating a `~/.aws/credentials` file. It is also possible to set the `endpoint_url` in the same file
to use a different S3 compatible service::

    [default]
    endpoint_url = https://some-storage.somewhere.world
    aws_access_key_id = xxxxxxxxxxxxxxxxxxxxxxxx
    aws_secret_access_key = xxxxxxxxxxxxxxxxxxxxxxxx

Alternatively, the `endpoint_url`, and keys can be set in one of
the `~/.config/anemoi/settings.toml`
or `~/.config/anemoi/settings-secrets.toml` files.
"""

import fnmatch
import logging
import os
import threading
from collections.abc import Iterable
from copy import deepcopy
from typing import Any

import tqdm

from ..config import load_config
from ..humanize import bytes_to_human
from . import BaseDownload
from . import BaseUpload

LOG = logging.getLogger(__name__)
SECRETS = ["aws_access_key_id", "aws_secret_access_key"]

# s3_clients are not thread-safe, so we need to create a new client for each thread

thread_local = threading.local()


def _s3_config(bucket: str, *, region: str = None) -> Any:
    """Get an S3 client config for the specified bucket and region.

    Parameters
    ----------
    bucket : str
        The name of the S3 bucket.
    region : str, optional
        The AWS region of the S3 bucket.

    Returns
    -------
    Any
        The S3 client.
    """
    from botocore import UNSIGNED

    boto3_config = {}

    if region:
        # This is using AWS

        options = {"region_name": region}

        # Anonymous access
        if not (
            os.path.exists(os.path.expanduser("~/.aws/credentials"))
            or ("AWS_ACCESS_KEY_ID" in os.environ and "AWS_SECRET_ACCESS_KEY" in os.environ)
        ):
            boto3_config["signature_version"] = UNSIGNED

    else:

        # We may be accessing a different S3 compatible service
        # Use anemoi.config to get the configuration

        region = "unknown-region"

        options = {"region_name": region}
        config = load_config(secrets=SECRETS)

        cfg = config.get("object-storage", {})
        candidate = None
        for k, v in cfg.items():
            if isinstance(v, (str, int, float, bool)):
                options[k] = v

            if isinstance(v, dict):
                if fnmatch.fnmatch(bucket, k):
                    if candidate is not None:
                        raise ValueError(f"Multiple object storage configurations match {bucket}: {candidate} and {k}")
                    candidate = k

        if candidate is not None:
            for k, v in cfg.get(candidate, {}).items():
                if isinstance(v, (str, int, float, bool)):
                    options[k] = v

        type = options.pop("type", "s3")
        if type != "s3":
            raise ValueError(f"Unsupported object storage type {type}")

        if "config" in options:
            boto3_config.update(options["config"])
            del options["config"]

    def _(options):

        def __(k, v):
            if k in SECRETS:
                return "***"
            return v

        if isinstance(options, dict):
            return {k: __(k, v) for k, v in options.items()}

        if isinstance(options, list):
            return [_(o) for o in options]

        return options

    LOG.debug(f"Using S3 options: {_(options)}")

    return boto3_config, options


def s3_options(bucket: str, *, region: str = None, service: str = "s3") -> dict:
    """Get the S3 configuration for the specified bucket and region.

    Parameters
    ----------
    bucket : str
        The name of the S3 bucket.
    region : str, optional
        The AWS region of the S3 bucket.
    service : str, optional
        The AWS service to use, default is "s3".

    Returns
    -------
    dict
        The S3 configuration.
    """
    _, options = _s3_config(bucket, region=region)
    return options


def s3_client(bucket: str, *, region: str = None, service: str = "s3") -> Any:
    """Get an S3 client for the specified bucket and region.

    Parameters
    ----------
    bucket : str
        The name of the S3 bucket.
    region : str, optional
        The AWS region of the S3 bucket.
    service : str, optional
        The AWS service to use, default is "s3".

    Returns
    -------
    Any
        The S3 client.
    """
    import boto3
    from botocore.client import Config

    if not hasattr(thread_local, "s3_clients"):
        thread_local.s3_clients = {}

    key = f"{bucket}-{region}-{service}"

    if key in thread_local.s3_clients:
        return thread_local.s3_clients[key]

    boto3_config, options = _s3_config(bucket, region=region)

    boto3_config.update(
        dict(
            max_pool_connections=25,
            request_checksum_calculation="when_required",
            response_checksum_validation="when_required",
        )
    )

    options["config"] = Config(**boto3_config)

    def _(options):

        def __(k, v):
            if k in SECRETS:
                return "***"
            return v

        if isinstance(options, dict):
            return {k: __(k, v) for k, v in options.items()}

        if isinstance(options, list):
            return [_(o) for o in options]

        return options

    LOG.debug(f"Using S3 options: {_(options)}")

    thread_local.s3_clients[key] = boto3.client(service, **options)

    return thread_local.s3_clients[key]


class S3Upload(BaseUpload):

    def get_temporary_target(self, target: str, pattern: str) -> str:
        """Get a temporary target path based on the given pattern.

        Parameters
        ----------
        target : str
            The original target path.
        pattern : str
            The pattern to format the temporary path.

        Returns
        -------
        str
            The temporary target path.
        """
        return target

    def rename_target(self, target: str, temporary_target: str) -> None:
        """Rename the target to a new target path.

        Parameters
        ----------
        target : str
            The original target path.
        temporary_target : str
            The new target path.
        """
        pass

    def delete_target(self, target: str) -> None:
        """Delete the target path.

        Parameters
        ----------
        target : str
            The target path to delete.
        """
        pass
        # delete(target)

    def _transfer_file(
        self,
        source: str,
        target: str,
        overwrite: bool,
        resume: bool,
        verbosity: int,
        threads: int,
        config: dict = None,
    ) -> int:
        """Transfer a file to S3.

        Parameters
        ----------
        source : str
            The source file path.
        target : str
            The target S3 path.
        overwrite : bool
            Whether to overwrite the target if it exists.
        resume : bool
            Whether to resume the transfer if possible.
        verbosity : int
            The verbosity level.
        threads : int
            The number of threads to use.
        config : dict, optional
            Additional configuration options.

        Returns
        -------
        int
            The size of the transferred file.

        Raises
        ------
        ValueError
            If the target already exists and overwrite or resume is not specified.
        """
        from botocore.exceptions import ClientError

        assert target.startswith("s3://")

        _, _, bucket, key = target.split("/", 3)
        s3 = s3_client(bucket)

        size = os.path.getsize(source)

        if verbosity > 0:
            LOG.info(f"{self.action} {source} to {target} ({bytes_to_human(size)})")

        try:
            results = s3.head_object(Bucket=bucket, Key=key)
            remote_size = int(results["ContentLength"])
        except ClientError as e:
            if e.response["Error"]["Code"] != "404":
                raise
            remote_size = None

        if remote_size is not None:
            if remote_size != size:
                LOG.warning(
                    f"{target} already exists, but with different size, re-uploading (remote={remote_size}, local={size})"
                )
            elif resume:
                # LOGGER.info(f"{target} already exists, skipping")
                return size

        if remote_size is not None and not overwrite and not resume:
            raise ValueError(f"{target} already exists, use 'overwrite' to replace or 'resume' to skip")

        if verbosity > 0:
            with tqdm.tqdm(total=size, unit="B", unit_scale=True, unit_divisor=1024, leave=False) as pbar:
                s3.upload_file(
                    source,
                    bucket,
                    key,
                    Callback=lambda x: pbar.update(x),
                    Config=config,
                )
        else:
            s3.upload_file(source, bucket, key, Config=config)

        return size


class S3Download(BaseDownload):

    def copy(self, source: str, target: str, **kwargs) -> None:
        """Copy a file or folder from S3 to the local filesystem.

        Parameters
        ----------
        source : str
            The source S3 path.
        target : str
            The target local path.
        kwargs : dict
            Additional arguments for the transfer.
        """
        assert source.startswith("s3://")

        if source.endswith("/"):
            self.transfer_folder(source=source, target=target, **kwargs)
        else:
            self.transfer_file(source=source, target=target, **kwargs)

    def list_source(self, source: str) -> Iterable:
        """List the objects in the source S3 path.

        Parameters
        ----------
        source : str
            The source S3 path.

        Returns
        -------
        Iterable
            An iterable of S3 objects.
        """
        yield from _list_objects(source)

    def source_path(self, s3_object: dict, source: str) -> str:
        """Get the S3 path of the object.

        Parameters
        ----------
        s3_object : dict
            The S3 object.
        source : str
            The source S3 path.

        Returns
        -------
        str
            The S3 path of the object.
        """
        _, _, bucket, _ = source.split("/", 3)
        return f"s3://{bucket}/{s3_object['Key']}"

    def target_path(self, s3_object: dict, source: str, target: str) -> str:
        """Get the local path for the S3 object.

        Parameters
        ----------
        s3_object : dict
            The S3 object.
        source : str
            The source S3 path.
        target : str
            The target local path.

        Returns
        -------
        str
            The local path for the S3 object.
        """
        _, _, _, folder = source.split("/", 3)
        local_path = os.path.join(target, os.path.relpath(s3_object["Key"], folder))
        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        return local_path

    def source_size(self, s3_object: dict) -> int:
        """Get the size of the S3 object.

        Parameters
        ----------
        s3_object : dict
            The S3 object.

        Returns
        -------
        int
            The size of the S3 object.
        """
        return s3_object["Size"]

    def _transfer_file(
        self,
        source: str,
        target: str,
        overwrite: bool,
        resume: bool,
        verbosity: int,
        threads: int,
        config: dict = None,
    ) -> int:
        """Transfer a file from S3 to the local filesystem.

        Parameters
        ----------
        source : str
            The source S3 path.
        target : str
            The target local path.
        overwrite : bool
            Whether to overwrite the target if it exists.
        resume : bool
            Whether to resume the transfer if possible.
        verbosity : int
            The verbosity level.
        threads : int
            The number of threads to use.
        config : dict, optional
            Additional configuration options.

        Returns
        -------
        int
            The size of the transferred file.

        Raises
        ------
        ValueError
            If the target does not exist on S3.
        """
        # from boto3.s3.transfer import TransferConfig

        _, _, bucket, key = source.split("/", 3)
        s3 = s3_client(bucket)

        if key.endswith("/"):
            return 0

        try:
            response = s3.head_object(Bucket=bucket, Key=key)
        except s3.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                raise ValueError(f"{source} does not exist ({bucket}, {key})")
            raise

        size = int(response["ContentLength"])

        if verbosity > 0:
            LOG.info(f"{self.action} {source} to {target} ({bytes_to_human(size)})")

        if overwrite:
            resume = False

        if resume:
            if os.path.exists(target):
                local_size = os.path.getsize(target)
                if local_size != size:
                    LOG.warning(
                        f"{target} already with different size, re-downloading (remote={size}, local={local_size})"
                    )
                else:
                    # if verbosity > 0:
                    #     LOGGER.info(f"{target} already exists, skipping")
                    return size

        if os.path.exists(target) and not overwrite:
            raise ValueError(f"{target} already exists, use 'overwrite' to replace or 'resume' to skip")

        if verbosity > 0:
            with tqdm.tqdm(total=size, unit="B", unit_scale=True, unit_divisor=1024, leave=False) as pbar:
                s3.download_file(
                    bucket,
                    key,
                    target,
                    Callback=lambda x: pbar.update(x),
                    Config=config,
                )
        else:
            s3.download_file(bucket, key, target, Config=config)

        return size


def _list_objects(target: str, batch: bool = False) -> Iterable:
    """List the objects in the target S3 path.

    Parameters
    ----------
    target : str
        The target S3 path.
    batch : bool, optional
        Whether to return objects in batches, by default False.

    Returns
    -------
    Iterable
        An iterable of S3 objects.
    """
    _, _, bucket, prefix = target.split("/", 3)
    s3 = s3_client(bucket)

    paginator = s3.get_paginator("list_objects_v2")

    for page in paginator.paginate(Bucket=bucket, Prefix=prefix):
        if "Contents" in page:
            objects = deepcopy(page["Contents"])
            if batch:
                yield objects
            else:
                yield from objects


def delete_folder(target: str) -> None:
    """Delete a folder from S3.

    Parameters
    ----------
    target : str
        The target S3 folder path.
    """
    _, _, bucket, _ = target.split("/", 3)
    s3 = s3_client(bucket)

    total = 0
    for batch in _list_objects(target, batch=True):
        LOG.info(f"Deleting {len(batch):,} objects from {target}")
        s3.delete_objects(Bucket=bucket, Delete={"Objects": [{"Key": o["Key"]} for o in batch]})
        total += len(batch)
        LOG.info(f"Deleted {len(batch):,} objects (total={total:,})")


def delete_file(target: str) -> None:
    """Delete a file from S3.

    Parameters
    ----------
    target : str
        The target S3 file path.
    """
    from botocore.exceptions import ClientError

    _, _, bucket, key = target.split("/", 3)
    s3 = s3_client(bucket)

    try:
        s3.head_object(Bucket=bucket, Key=key)
        exits = True
    except ClientError as e:
        if e.response["Error"]["Code"] != "404":
            raise
        exits = False

    if not exits:
        LOG.warning(f"{target} does not exist. Did you mean to delete a folder? Then add a trailing '/'")
        return

    LOG.info(f"Deleting {target}")
    s3.delete_object(Bucket=bucket, Key=key)
    LOG.info(f"{target} is deleted")


def delete(target: str) -> None:
    """Delete a file or a folder from S3.

    Parameters
    ----------
    target : str
        The URL of a file or a folder on S3. The URL should start with 's3://'.
    """

    assert target.startswith("s3://")

    if target.endswith("/"):
        delete_folder(target)
    else:
        delete_file(target)


def list_folder(folder: str) -> Iterable:
    """List the subfolders in a folder on S3.

    Parameters
    ----------
    folder : str
        The URL of a folder on S3. The URL should start with 's3://'.

    Returns
    -------
    list
        A list of the subfolder names in the folder.
    """

    assert folder.startswith("s3://")
    if not folder.endswith("/"):
        folder += "/"

    _, _, bucket, prefix = folder.split("/", 3)

    s3 = s3_client(bucket)
    paginator = s3.get_paginator("list_objects_v2")

    for page in paginator.paginate(Bucket=bucket, Prefix=prefix, Delimiter="/"):
        if "CommonPrefixes" in page:
            yield from [folder + _["Prefix"] for _ in page.get("CommonPrefixes") if _["Prefix"] != "/"]
        if "Contents" in page:
            yield from [folder + _["Key"] for _ in page.get("Contents")]


def object_info(target: str) -> dict:
    """Get information about an object on S3.

    Parameters
    ----------
    target : str
        The URL of a file or a folder on S3. The URL should start with 's3://'.

    Returns
    -------
    dict
        A dictionary with information about the object.
    """

    _, _, bucket, key = target.split("/", 3)
    s3 = s3_client(bucket)

    try:
        return s3.head_object(Bucket=bucket, Key=key)
    except s3.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "404":
            raise FileNotFoundError(f"{target} does not exist")
        raise


def object_exists(target: str) -> bool:
    """Check if an object exists.

    Parameters
    ----------
    target : str
        The URL of a file or a folder on S3. The URL should start with 's3://'.

    Returns
    -------
    bool
        True if the object exists, False otherwise.
    """

    _, _, bucket, key = target.split("/", 3)
    s3 = s3_client(bucket)

    try:
        s3.head_object(Bucket=bucket, Key=key)
        return True
    except s3.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "404":
            return False
        raise


def object_acl(target: str) -> dict:
    """Get information about an object's ACL on S3.

    Parameters
    ----------
    target : str
        The URL of a file or a folder on S3. The URL should start with 's3://'.

    Returns
    -------
    dict
        A dictionary with information about the object's ACL.
    """

    _, _, bucket, key = target.split("/", 3)
    s3 = s3_client(bucket)

    return s3.get_object_acl(Bucket=bucket, Key=key)


def download(source: str, target: str, *args, **kwargs) -> None:
    """Download a file or folder from S3 to the local filesystem.

    Parameters
    ----------
    source : str
        The source S3 path.
    target : str
        The target local path.
    args : tuple
        Additional positional arguments.
    kwargs : dict
        Additional keyword arguments.
    """
    from . import transfer

    assert source.startswith("s3://"), f"source {source} should start with 's3://'"
    return transfer(source, target, *args, **kwargs)


def upload(source: str, target: str, *args, **kwargs) -> None:
    """Upload a file or folder to S3.

    Parameters
    ----------
    source : str
        The source file or folder path.
    target : str
        The target S3 path.
    args : tuple
        Additional positional arguments.
    kwargs : dict
        Additional keyword arguments.
    """
    from . import transfer

    assert target.startswith("s3://"), f"target {target} should start with 's3://'"
    return transfer(source, target, *args, **kwargs)


def quotas(target: str) -> dict:
    """Get the quotas for an S3 bucket.

    Parameters
    ----------
    target : str
        The URL of a file or a folder on S3. The URL should start with 's3://'.

    Returns
    -------
    dict
        A dictionary with the quotas for the bucket.
    """
    from botocore.exceptions import ClientError

    _, _, bucket, _ = target.split("/", 3)
    s3 = s3_client(bucket, service="service-quotas")

    try:
        return s3.list_service_quotas(ServiceCode="ec2")
    except ClientError as e:
        if e.response["Error"]["Code"] == "404":
            raise ValueError(f"{target} does not exist")
        raise



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import os
import re
from pathlib import Path
from typing import Any
from urllib.parse import parse_qs
from urllib.parse import urlencode
from urllib.parse import urlparse
from urllib.parse import urlunparse

# Patterns used by earthkit-data for url-patterns and path-patterns

RE1 = re.compile(r"{([^}]*)}")  # {*}
RE2 = re.compile(r"\(([^}]*)\)")  # (*)


def sanitise(obj: Any, level=1) -> Any:
    """Sanitise an object by replacing all full paths with shortened versions and URL credentials with '***'.

    Parameters
    ----------
    obj : Any
        The object to sanitise.
    level : int, optional
        The level of sanitation. The higher levels will also apply the levels below it.
        - 1: Shorten file paths to file name and hide credentials in URLs (default).
        - 2: Hide hostnames in URLs.
        - 3: Hide full file paths and URLs.

    Returns
    -------
    Any
        The sanitised object.
    """

    assert level in (1, 2, 3), "level must be 1, 2 or 3"

    if isinstance(obj, dict):
        return {sanitise(k): sanitise(v) for k, v in obj.items()}

    if isinstance(obj, list):
        return [sanitise(v) for v in obj]

    if isinstance(obj, tuple):
        return tuple(sanitise(v) for v in obj)

    if isinstance(obj, str):
        return _sanitise_string(obj, level)

    return obj


def _sanitise_string(obj: str, level=1) -> str:
    """Sanitise a string by replacing full paths and URL passwords."""

    parsed = urlparse(obj, allow_fragments=True)

    if parsed.scheme and parsed.scheme[0].isalpha():
        return _sanitise_url(parsed, level)

    if level > 2:
        return "hidden"

    if obj.startswith("/") or obj.startswith("~"):
        return _sanitise_path(obj)

    return obj


def _sanitise_url(parsed: Any, level=1) -> str:
    """Sanitise a URL by replacing passwords with '***'."""

    LIST = [
        "pass",
        "password",
        "token",
        "user",
        "key",
        "pwd",
        "_key",
        "_token",
        "apikey",
        "api_key",
        "api_token",
        "_api_token",
        "_api_key",
        "username",
        "login",
        "auth",
        "auth_token",
        "auth_key",
    ]

    scheme, netloc, path, params, query, fragment = parsed

    if parsed.password or parsed.username:
        _, host = netloc.split("@")
        user = "user:***" if parsed.password else "user"
        netloc = f"{user}@{host}"

    if query:
        qs = parse_qs(query)
        for k in LIST:
            if k in qs:
                qs[k] = "hidden"
        query = urlencode(qs, doseq=True)

    if params:
        qs = parse_qs(params)
        for k in LIST:
            if k in qs:
                qs[k] = "hidden"
        params = urlencode(qs, doseq=True)

    if level > 1:
        if (bits := netloc.split("@")) and len(bits) > 1:
            netloc = f"{bits[0]}@hidden"
        else:
            netloc = "hidden"

    if level > 2:
        return urlunparse([scheme, netloc, "", "", "", ""])

    return urlunparse([scheme, netloc, path, params, query, fragment])


def _sanitise_path(path: str) -> str:
    """Sanitise a file path by shortening it."""
    bits = list(reversed(Path(path).parts))
    result = [bits.pop(0)]
    for bit in bits:
        if RE1.match(bit) or RE2.match(bit):
            # keep earthkit-data folder patterns
            result.append(bit)
            continue
        if result[-1] == "...":
            continue
        result.append("...")
    result = os.path.join(*reversed(result))
    if bits[-1] == "/":
        result = os.path.join("/", result)

    return result
